#include "Lgi.h"
#include "Ftp.h"
#include "GTabView.h"
#include "GCombo.h"
#include "GCheckBox.h"
#include "GEdit.h"

char *TempStr = "(temporary)";

ConnectDlg::ConnectDlg(AppWnd *parent, List<FtpAcc> *accounts)
{
	SetParent(App = parent);
	Accounts = accounts;
	Account = 0;
	Index = 0;
	Cur = 0;

	Fields[(char*)OPT_RemoteHost] = IDC_SERVER;
	Fields[(char*)OPT_Port] = IDC_PORT;
	Fields[(char*)OPT_UserName] = IDC_USERNAME;
	Fields[(char*)OPT_Password] = IDC_PASSWORD;
	Fields[(char*)OPT_Anonymous] = IDC_ANONYMOUS;
	Fields[(char*)OPT_LocalDir] = IDC_LOCAL;
	Fields[(char*)OPT_RemoteDir] = IDC_REMOTE;
	Fields[(char*)OPT_UseActive] = IDC_USE_ACTIVE;
	Fields[(char*)OPT_NormalList] = IDC_SIMPLE_LIST;
	Fields[(char*)OPT_SFtp] = IDC_SFTP;

	if (parent)
	{
		OptTools.Set(this, parent->GetOptions());
		OptFields.Insert(NEW(DataDlgField(DATA_STR, IDC_SOCKS5_SERVER, OPT_Socks5Server)));
		OptFields.Insert(NEW(DataDlgField(DATA_STR, IDC_SOCKS5_USERNAME, OPT_Socks5UserName)));
		OptFields.Insert(NEW(DataDlgField(DATA_STR, IDC_SOCKS5_PASSWORD, OPT_Socks5Password)));
	}

	if (LoadFromResource(IDD_CONNECT))
	{
		GTabView *Tab = dynamic_cast<GTabView*>(FindControl(IDC_TAB));
		if (Tab)
		{
			MoveToCenter();

			Cur = Temp = NEW(FtpAcc);
			if (Temp)
			{
				GVariant v = TempStr;
				Temp->SetVariant(OPT_AccountName, v);
				v = 21;
				Temp->SetVariant(OPT_Port, v);

				Accounts->Insert(Temp, 0);
			}

			AccountCombo = dynamic_cast<GCombo*>(FindControl(IDC_ACCOUNT));
			LoadAccountList();

			// Data -> Controls
			Cur->SerializeUI(this, Fields, true);
			
			// Load options
			OptTools.ProcessFields(OptFields, false);

			DoModal();
		}
	}
}

ConnectDlg::~ConnectDlg()
{
	Accounts->Delete(Temp);
	DeleteObj(Temp);
}

FtpAcc *ConnectDlg::GetAccount(char *Name)
{
	if (Name)
	{
		for (FtpAcc *a = Accounts->First(); a; a = Accounts->Next())
		{
			GVariant n;
			if (a->GetVariant(OPT_AccountName, n) AND
				n.Str() AND
				stricmp(Name, n.Str()) == 0)
			{
				return a;
			}
		}
	}
	return 0;
}

void ConnectDlg::LoadAccountList()
{
	if (AccountCombo AND Accounts)
	{
		while (AccountCombo->Delete(0));

		for (FtpAcc *a = Accounts->First(); a; a = Accounts->Next())
		{
			GVariant n;
			if (a->GetVariant(OPT_AccountName, n) AND n.Str())
			{
				AccountCombo->Insert(n.Str());
			}
			else
			{
				Accounts->Delete(a);
				DeleteObj(a);
			}
		}

		AccountCombo->Value(Index);
	}
}

void ConnectDlg::Serialize(bool Write)
{
	if (Write)
	{
		GCheckBox *Anon = dynamic_cast<GCheckBox*>(FindControl(IDC_ANONYMOUS));

		// check for URL and decode
		char *Protocol = "ftp://";

		char *Server = NewStr(GetCtrlName(IDC_SERVER));
		if (Server AND strnicmp(Server, Protocol, strlen(Protocol)) == 0)
		{
			// ftp:// URL
			char *s = Server + strlen(Protocol);
			
			// User/Password
			char *At = strchr(s, '@');
			if (At)
			{
				*At = 0;
				char *Colon = strchr(s, ':');
				if (Colon)
				{
					*Colon = 0;
					SetCtrlName(IDC_USERNAME, s);
					SetCtrlName(IDC_PASSWORD, Colon+1);
				}
				s = At + 1;

				if (Anon)
				{
					Anon->Value(false);
				}
			}
			else if (Anon)
			{
				Anon->Value(true);
			}

			// Remote path
			char *Dir = strchr(s, '/');
			if (Dir)
			{
				// Chop off any file name
				char *LastDir = 0;
				for (char *End = Dir+strlen(Dir)-1; End>Dir; End--)
				{
					if (*End == '/')
					{
						LastDir = End;
						break;
					}
				}
				if (LastDir AND strchr(LastDir, '.'))
				{
					// er it's a file name... choppy chop
					*LastDir = 0;
				}

				SetCtrlName(IDC_REMOTE, Dir);
				*Dir++ = 0;
			}

			// Port?
			char *Colon = strchr(s, ':');
			if (Colon)
			{
				*Colon++ = 0;
				SetCtrlName(IDC_PORT, Colon);
			}

			// All thats left is the remote host
			SetCtrlName(IDC_SERVER, s);
		}
		DeleteArray(Server);

		// Check for user / anon
		if (Anon AND NOT ValidStr(GetCtrlName(IDC_USERNAME)))
		{
			Anon->Value(true);
		}

		// Data -> Controls
		Cur->SerializeUI(this, Fields, false);
	}
	else
	{
		// Controls -> Data
		Cur->SerializeUI(this, Fields, true);
	}
}

int ConnectDlg::OnNotify(GView *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		case IDC_USERNAME:
		{
			SetCtrlValue(IDC_ANONYMOUS, NOT ValidStr(Ctrl->Name()));
			break;
		}
		case IDC_ACCOUNT:
		{
			if (AccountCombo)
			{
				int i = AccountCombo->Value();
				if (i >= 0 AND Index != i)
				{
					Index = i;
					FtpAcc *Acc = Accounts->ItemAt(Index);
					if (Acc)
					{
						Serialize(true);
						Cur = Acc;
						Serialize(false);
						AccountCombo->Value(Index);
					}
				}
				// else user is typing something
			}
			break;
		}
		case IDC_ADD:
		{
			if (AccountCombo)
			{
				// Controls -> Data
				Serialize(true);

				FtpAcc *Acc = 0;
				if (Cur)
				{
					// save the temp into the list
					Acc = NEW(FtpAcc);
					if (Acc)
					{
						// copy to a new object
						*Acc = *Temp;

						// Work out the account name
						char AName[256] = "";
						GInput Dlg(this, AName, LgiLoadString(IDS_NEW_ACCOUNT_NAME), AppName);
						if (Dlg.DoModal() AND Dlg.Str)
						{
							strcpy(AName, Dlg.Str);
						}
						else return 0;
						
						GVariant v(AName);
						Acc->SetVariant(OPT_AccountName, v);

						// new current record
						// Options = Acc;
						Accounts->Insert(Cur = Acc);
						Index = Accounts->IndexOf(Acc);

						// delete keys out of temp... starts blank
						Temp->Empty();
						v = TempStr;
						Temp->SetVariant(OPT_AccountName, v);
						v = 21;
						Temp->SetVariant(OPT_Port, v);

						// load the data back into the UI
						// Data -> Controls
						Serialize(false);
					}
				}

				LoadAccountList();
			}
			break;
		}
		case IDC_DELETE:
		{
			if (AccountCombo AND Index > 0)
			{
				Accounts->Delete(Cur);
				Cur = Accounts->ItemAt(--Index);
				Serialize(false); // Data -> Controls
				LoadAccountList();
			}
			break;
		}
		case IDC_BROWSE:
		{
			GEdit *LocalDir = dynamic_cast<GEdit*>(FindControl(IDC_LOCAL));
			if (LocalDir)
			{
				GFileSelect Select;
	
				Select.Name(LocalDir->Name());
				Select.Parent(this);
				if (Select.OpenFolder())
				{
					LocalDir->Name(Select.Name());
				}
			}
			break;
		}
		case IDOK:
		{
			// Save connect fields
			Serialize(true);
			Account = Cur;

			// Save options
			OptTools.ProcessFields(OptFields, true);

			EndModal(1);
			break;
		}
		case IDCANCEL:
		{
			Cur->SerializeUI(this, Fields, false);

			// Save options
			OptTools.ProcessFields(OptFields, true);

			EndModal(0);
			break;
		}
	}
	return 0;
}

