/*
**	FILE:			FtpApp.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			7/9/99
**	DESCRIPTION:	i.Ftp application
**
**	Copyright (C) 1999, Matthew Allen
**		fret@memecode.com
*/

// Debug defines

// Includes
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <stdarg.h>
#include <ctype.h>

#include "Ftp.h"
#include "ISFtp.h"
#include "GToken.h"
#include "IHttp.h"
#include "GAbout.h"
#include "resdefs.h"
#include "GArray.h"
#include "GDropFiles.h"
#include "GCheckBox.h"
#include "GEdit.h"
#include "GTextLabel.h"
#include "GButton.h"
#include "GCombo.h"
#include "Schedule.h"

// #define IDM_HTTP				10001

#define IDLE_NOOP_TIME			60

#ifdef BEOS
#define isdigit(c) ((c) >= '0' AND (c) <= '9')
#endif

#define FTP_SAVE_FILE			(M_USER+1202)

#define IDM_LANG_BASE			1000

//////////////////////////////////////////////////////////////////////////////
char AppName[] = "i.Ftp";
char HelpFile[] = "index.html";
char OptionsFileName[] = "options.sgml";

//////////////////////////////////////////////////////////////////////////////
int Check(IFtpEntry *n)
{
	if (n == 0 OR (int)(n->Name) < 100)
	{
		printf("Check(IFtpEntry *%p) failed!!!\n", n);
		return false;
	}
	
	return true;
}

int Check(List<IFtpEntry> &n)
{
	Iterator<IFtpEntry> i(&n);
	for (IFtpEntry *e=i.First(); e; e=i.Next())
	{
		if (NOT Check(e))
		{
			printf("Check(List<IFtpEntry>&) failed!!!\n");
			return false;
		}
	}
	
	return true;
}

//////////////////////////////////////////////////////////////////////////////
void GetFullAppName(char *s)
{
	if (s)
	{
		#ifndef _DEBUG
		char *Build = "Release";
		#else
		char *Build = "Debug";
		#endif

		sprintf(s, "%s v%02.2f (%s %s)", AppName, VER, LgiGetOsName(), Build);
	}
}

void Error(char *str, ...)
{
	if (str)
	{
		char buffer[256];
		va_list arg;
		va_start(arg ,str);
		vsprintf(buffer, str, arg);
		va_end(arg);

		LgiMsg(MainWnd, buffer, LgiLoadString(IDS_ERROR), MB_OK);
	}
}

//////////////////////////////////////////////////////////////////////////////
void ConvertUrlToAccount(char *Url, FtpAcc *Account, char **FileName = 0)
{
	if (Url AND Account)
	{
		// check for URL and decode
		char *Protocol = "ftp://";

		if (strnicmp(Url, Protocol, strlen(Protocol)) == 0)
		{
			// ftp:// URL
			char *Buf = NewStr(Url + strlen(Protocol));
			char *s = Buf;
			
			// User/Password
			GVariant v((int)1);
			Account->SetVariant(OPT_Anonymous, v);
			char *At = strchr(s, '@');
			if (At)
			{
				*At = 0;
				char *Colon = strchr(s, ':');
				if (Colon)
				{
					*Colon = 0;

					v = s;
					Account->SetVariant(OPT_UserName, v);
					v = Colon+1;
					Account->SetVariant(OPT_Password, v);
					v = (int)0;
					Account->SetVariant(OPT_Anonymous, v);
				}
				s = At + 1;
			}

			// Remote path
			char *Dir = strchr(s, '/');
			if (Dir)
			{
				// Chop off any file name
				char *LastDir = strrchr(Dir, '/');
				if (LastDir AND strchr(LastDir, '.'))
				{
					if (FileName)
					{
						*FileName = NewStr(LastDir + 1);
					}

					// er it's a file name... choppy chop
					*LastDir = 0;
				}

				if (strlen(Dir) > 0)
				{
					v = Dir + 1;
					Account->SetVariant(OPT_RemoteDir, v);
				}
				*Dir++ = 0;
			}

			// Port?
			char *Colon = strchr(s, ':');
			if (Colon)
			{
				*Colon++ = 0;
				v = atoi(Colon);
				Account->SetVariant(OPT_Port, v);
			}

			// All thats left is the remote host
			v = s;
			Account->SetVariant(OPT_RemoteHost, v);

			DeleteArray(Buf);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
class OptionsDlg : public GDialog, public DataDlgTools
{
	List<DataDlgField> Feilds;

public:
	OptionsDlg(GView *Parent, ObjProperties *o)
	{
		Options = o;
		SetParent(Parent);
		Dlg = this;
		DeleteEmptyStrings = true;

		if (LoadFromResource(IDD_OPTIONS))
		{
			Feilds.Insert(NEW(DataDlgField(DATA_BOOL, IDC_CONFIRM_DEL, OPT_ConfirmDeletes)));
			Feilds.Insert(NEW(DataDlgField(DATA_BOOL, IDC_CONFIRM_LINKS, OPT_ConfirmLinks)));
			Feilds.Insert(NEW(DataDlgField(DATA_BOOL, IDC_RESIZE_TO_CONTENT, OPT_ResizeToContent)));
			Feilds.Insert(NEW(DataDlgField(DATA_STR, IDC_ASCII_TYPES, OPT_AsciiTypes)));
			Feilds.Insert(NEW(DataDlgField(DATA_STR, IDC_BINARY_TYPES, OPT_BinaryTypes)));
			Feilds.Insert(NEW(DataDlgField(DATA_STR, IDC_LOG, OPT_LogFile)));
			Feilds.Insert(NEW(DataDlgField(DATA_INT, IDC_DBL_CLICK, OPT_DblClickAction)));

			MoveToCenter();
			LoadFields(Feilds);
			DoModal();
		}
	}

	int OnNotify(GView *c, int f)
	{
		switch (c->GetId())
		{
			case IDOK:
			{
				SaveFields(Feilds);
				// fall thru
			}
			case IDCANCEL:
			{
				EndModal(c->GetId() == IDOK);
				break;
			}
		}

		return 0;
	}
};

//////////////////////////////////////////////////////////////////////////////
class AppWndPrivate
{
public:
	GXmlTag *Schedule;

	AppWndPrivate()
	{
		Schedule = 0;
	}

	~AppWndPrivate()
	{
		DeleteObj(Schedule);
	}
};

AppWnd::AppWnd()
{
	// init some variables
	d = NEW(AppWndPrivate);
	GApp::ObjInstance()->AppWnd = this;
	ImageList = 0;
	Splitter = 0;
	OptionsFile = 0;
	Connection = 0;
	Ftp = 0;
	ImageList = 0;
	MaxDataRate = -1;
	AccountLst = 0;
	Local = 0;
	LastCmd = 0;
	BinaryMode = true;
	ZeroObj(Modes);
	Http = 0;
	Worker = 0;
	Restart = false;

	// Default position
	GRect r(50, 50, 870, 550);
	SetPos(r);

	// Load the settings
	Serialize(false);

	char *LangId = 0;
	if (Options.Get(OPT_Language, LangId))
	{
		// Set the language to load...
		XmlTag *t = NEW(XmlTag);
		if (t)
		{
			t->Name = "language";
			t->Set("id", LangId);
			LgiApp->SetConfig(t);
		}
	}

	int Resize;
	if (NOT GetOptions()->Get(OPT_ResizeToContent, Resize))
	{
		GetOptions()->Set(OPT_ResizeToContent, true);
	}

	Worker = NEW(FtpWorkerThread);

	// Startup UI
	Name(AppName);
	SetQuitOnClose(true);

	#if defined WIN32
	CreateClass(AppName, LoadIcon(LgiProcessInst(), MAKEINTRESOURCE(IDI_ICON1)));
	#endif

	if (NOT Attach(0))
	{
		LgiMsg(0, LgiLoadString(IDS_ERR_MAINWND));
		LgiExitApp();
	}
	else
	{
		SetupUi();
		SetWindow(this);
		SetQuitOnClose(true);
		
		#if defined LINUX
		char *Icon = LgiFindFile("_icon.png");
		if (Icon)
		{
			Handle()->setIcon(Icon);
			DeleteArray(Icon);
		}
		#endif
	}

	if (Local)
	{
		char *Folder = 0;
		if (GetOptions()->Get(OPT_LocalFolder, Folder))
		{
			if (*Folder == '.')
			{
				char p[300];
				LgiGetExePath(p, sizeof(p));
				LgiMakePath(p, sizeof(p), p, Folder);
				Local->Cd(p);
			}
			else
			{
				Local->Cd(Folder);
			}
		}
		else
		{
			char Exe[300];
			if (LgiGetExePath(Exe, sizeof(Exe)))
			{
				Local->Cd(Exe);
			}
		}
	}

	DoCmdLine();

	char ScheduleFile[300];
	if (GetScheduleFile(ScheduleFile, sizeof(ScheduleFile)) AND
		FileExists(ScheduleFile))
	{
		OnScheduleChanged(ScheduleFile);
	}
}

AppWnd::~AppWnd()
{
	char *Path = Local->Edit->Name();
	if (ValidStr(Path))
	{
		char Exe[300];
		LgiGetExePath(Exe, sizeof(Exe));
		char *Rel = LgiMakeRelitivePath(Exe, Path);
		if (Rel)
		{
			GetOptions()->Set(OPT_LocalFolder, Rel);
			DeleteArray(Rel);
		}
		else
		{		
			GetOptions()->Set(OPT_LocalFolder, Path);
		}
	}

	if (LogPanel)
	{
		GetOptions()->Set(OPT_LogOpen, LogPanel->Open());
	}
	LgiApp->AppWnd = 0;
	
	Options.Set("binary", BinaryMode);
	if (Splitter)
	{
		Options.Set("split-pos", Splitter->Value());
	}

	Serialize(true);
	
	if (Ftp)
	{
		Ftp->Close();
	}

	DeleteArray(OptionsFile);
	DeleteObj(ImageList);
	DeleteObj(Worker);
	DeleteObj(Progress);
	DeleteObj(Ftp);
	Accounts.DeleteObjects();
	DeleteObj(d);
	
	if (Restart)
	{
		char Exe[256];
		LgiGetExeFile(Exe, sizeof(Exe));
		printf("Exe='%s'\n", Exe);
		LgiExecute(Exe);
		LgiSleep(1000);
	}
}

bool AppWnd::ShowHiddenFiles()
{
	int ShowHidden = false;
	Options.Get(OPT_ShowHidden, ShowHidden);
	return ShowHidden;
}

FtpAcc *AppWnd::ArgToAccount(char *Arg, char **File)
{
	FtpAcc *a = 0;

	if (Arg)
	{
		if (stristr(Arg, "ftp://"))
		{
			a = NEW(FtpAcc);
			if (a)
			{
				ConvertUrlToAccount(Arg, a, File);
			}
		}
		else
		{
			for (FtpAcc *Acc = Accounts.First(); Acc; Acc = Accounts.Next())
			{
				GVariant v;
				if (Acc->GetVariant(OPT_AccountName, v) AND
					v.Str() AND
					stricmp(v.Str(), Arg) == 0)
				{
					a = NEW(FtpAcc);
					if (a)
					{
						*a = *Acc;
						break;
					}
				}
			}
		}
	}

	return a;
}

void AppWnd::DoCmdLine()
{
	char Src[256];
	char Dest[256];
	bool ExitWhenDone = GApp::ObjInstance()->GetOption("x", Src);
	bool Overwrite = GApp::ObjInstance()->GetOption("over", Src);

	if (GApp::ObjInstance()->GetOption("s", Src) AND
		GApp::ObjInstance()->GetOption("d", Dest))
	{
		FtpAcc *Account = 0;
		char *File = 0;

		if (Account = ArgToAccount(Src, &File))
		{
			// Download
			ConnectTo(Account);

			// Setup local directory...
			Local->Cd(Dest);
			Local->ListElements();

			// Transfer pattern...
			FtpCmd *Trans = NEW(FtpCmd(File));
			if (Trans)
			{
				Trans->Cmd = TASK_TransferPattern;
				Trans->View = Remote;
				Trans->Int2 = Overwrite;
				Trans->Connection = Connection;

				if (ExitWhenDone)
				{
					// Exit on done...
					FtpCmd *Exit;
					Trans->OnSuccess.Insert(Exit = NEW(FtpCmd));
					if (Exit)
					{
						Exit->Cmd = TASK_Exit;
						Exit->View = Remote;
						Exit->Connection = Connection;
					}
				}
				else
				{
					// List the local contents
					FtpCmd *Lst;
					Trans->OnSuccess.Insert(Lst = NEW(FtpCmd));
					if (Lst)
					{
						Lst->Cmd = TASK_ListElements;
						Lst->View = Local;
						Lst->Connection = Connection;
					}
				}

				Remote->PostCmd(Trans);
			}
		}
		else if (Account = ArgToAccount(Dest, &File))
		{
			// Upload
			ConnectTo(Account);

			// Post Upload Cmd
			FtpCmd *Cmd = NEW(FtpCmd);
			if (Cmd)
			{
				Cmd->Cmd = TASK_Transfer;
				Cmd->View = Local;
				Cmd->Connection = Connection;
				strcpy(Cmd->Arg, Src);

				// Post List Cmd
				FtpCmd *Lst;
				Cmd->OnSuccess.Insert(Lst = NEW(FtpCmd));
				if (Lst)
				{
					Lst->Cmd = TASK_ListElements;
					Lst->View = Remote;
					Lst->Connection = Connection;

					if (ExitWhenDone)
					{
						// Exit on done...
						FtpCmd *Exit;
						Lst->OnSuccess.Insert(Exit = NEW(FtpCmd));
						if (Exit)
						{
							Exit->Cmd = TASK_Exit;
							Exit->View = Remote;
							Exit->Connection = Connection;
						}
					}
				}

				Local->PostCmd(Cmd);
			}
		}

		DeleteArray(File);
	}
}	

void AppWnd::OnPulse()
{
	// printf("AppWnd::OnPulse start\n");

	if (Ftp AND Ftp->IsOpen())
	{
		int Now = LgiCurrentTime();
		int Diff = Now - LastCmd;
		bool Running = Worker ? Worker->IsWorking() : false;

		if (LastCmd AND
			Remote AND
			NOT Running AND
			Diff > (IDLE_NOOP_TIME * 1000))
		{
			FtpCmd *Cmd = NEW(FtpCmd);
			if (Cmd)
			{
				Cmd->Cmd = TASK_Noop;
				Cmd->View = Remote;
				Remote->PostCmd(Cmd);
			}
		}

		if (StatusInfo[0])
		{
			char Str[256] = "";
			if (Running)
			{
				sprintf(Str, "%s (%s: %i)", LgiLoadString(IDS_BUSY), LgiLoadString(IDS_QUEUE), Worker?Worker->GetCmdsLeft()+1:0);
			}
			else
			{
				int Seconds = (LgiCurrentTime() - LastCmd) / 1000;
				sprintf(Str, "%s: %i:%02.2i", LgiLoadString(IDS_IDLE), Seconds/60, Seconds%60);
			}
			StatusInfo[0]->Name(Str);
		}
	}

	if (d->Schedule AND NOT Ftp->IsOpen())
	{
		for (GXmlTag *t = d->Schedule->Children.First(); t; t = d->Schedule->Children.Next())
		{
			char *Done = t->GetAttr(OPT_Done);
			if (NOT Done OR NOT atoi(Done))
			{
				char *Time = t->GetAttr(OPT_Time);
				if (Time)
				{
					GDateTime Start, Now;
					Start.Set(Time);
					Now.SetNow();
					if (Start < Now)
					{
						t->SetAttr(OPT_Done, "1");

						char *Url = t->GetAttr(OPT_Url);
						char *Folder = t->GetAttr(OPT_Folder);
						if (Url)
						{
							// Ftp URL
							FtpAcc Account;
							char *File = 0;
							ConvertUrlToAccount(Url, &Account, &File);
							ConnectTo(&Account);
							
							if (File)
							{
								PostEvent(FTP_SAVE_FILE, (int)NewStr(Folder), (int)NewStr(File));
							}
						}
					}
				}
			}
		}
	}
}

void AppWnd::SetDataRate(int Max)
{
	MaxDataRate = Max;
}

int AppWnd::WritePipe(GStream *c, char *Data, int Len, int Flags)
{
	if (MaxDataRate >= 0)
	{
		int InitialLen = Len;
		while (Len > 0)
		{
			int Now = LgiCurrentTime();
			int SecondLen = min(Len, MaxDataRate);
			int Written = c->Write(Data, SecondLen, Flags);
			if (Written > 0)
			{
				Data += Written;
				Len -= Written;
				while (LgiCurrentTime() - Now < 1000)
				{
					// wait till next second
				}
			}
			else
			{
				break;
			}
		}

		return InitialLen - Len;
	}
	else
	{
		// Push it through as fast as we can
		return c->Write(Data, Len, Flags);
	}

	return 0;
}

int AppWnd::ReadPipe(GStream *c, char *Data, int Len, int Flags)
{
	int Timeout = 30 * 1000;
	int64 Start = LgiCurrentTime();

	GSocket *Sock = dynamic_cast<GSocket*>(c);
	if (MaxDataRate >= 0)
	{
		int Read = 0;

		if (Len > MaxDataRate)
		{
			Len = MaxDataRate;
		}

		while (Len > 0)
		{
			int r = 0;
			int m = min(Len, MaxDataRate);
			
			if (NOT Sock OR Sock->IsReadable())
			{
				r = c->Read(Data, m, Flags);
			}
			else if (LgiCurrentTime() - Start < Timeout)
			{
				LgiSleep(1);
				continue;
			}
			else
			{
				LgiTrace("ReadPipe timed out (%is)\n", Timeout / 1000);
				return -1;
			}

			if (r > 0)
			{
				Data += r;
				Len -= r;
				Read += r;
			}
			else
			{
				break;
			}
		}

		while (LgiCurrentTime() < Start + 1000)
		{
			// wait till next second
			LgiSleep(0);
		}

		return Read;
	}
	else
	{
		// Wait for data
		while (Sock AND NOT Sock->IsReadable())
		{
			LgiSleep(1);

			if (LgiCurrentTime() - Start > Timeout)
			{
				LgiTrace("ReadPipe timed out (%is)\n", Timeout / 1000);
				return -1;
			}
		}

		// Read it in as fast as we can
		return c->Read(Data, Len, Flags);
	}

	return 0;
}

bool AppWnd::OnRequestClose(bool OsShuttingDown)
{
	if (Worker)
	{
		// Ask the worker to stop
		Worker->SetLoop(false);

		// Wait a bit
		int Start = LgiCurrentTime();
		while (	NOT Worker->IsExited() AND
				LgiCurrentTime() < Start + 2000)
		{
			LgiSleep(100);
		}

		if (NOT Worker->IsExited())
		{
			if (Connection)
			{
				// Close the socket from underneath the worker
				Connection->Close();
			}

			// Give it some more time to finish
			Start = LgiCurrentTime();
			while (	NOT Worker->IsExited() AND
					LgiCurrentTime() < Start + 1000)
			{
				LgiSleep(100);
			}

			if (NOT Worker->IsExited())
			{
				// Alright times up... DIE!
				Worker->Terminate();
			}
		}
	}

	DeleteObj(Http);

	return GWindow::OnRequestClose(OsShuttingDown);
}

void AppWnd::SetupUi()
{
	Menu = NEW(GMenu);
	if (Menu AND
		Menu->Attach(this) AND
		Menu->Load(this, "IDM_MENU"))
	{
		CmdConnect.MenuItem = Menu->FindItem(IDM_CONNECT);
		CmdDisconnect.MenuItem = Menu->FindItem(IDM_DISCONNECT);
		CmdHttp.MenuItem = Menu->FindItem(IDM_HTTP);
		
		GMenuItem *i = Menu->FindItem(IDM_NOOP);
		if (i)
		{
			AccountLst = i->GetParent();
		}

		Modes[0] = Menu->FindItem(IDM_ASCII);
		Modes[1] = Menu->FindItem(IDM_BINARY);

		int Bin;
		if (Options.Get("Binary", Bin))
		{
			BinaryMode = Bin;
		}

		Modes[BinaryMode ? 1 : 0]->Checked(true);
		
		GSubMenu *Lang = Menu->FindSubMenu(IDM_LANGUAGE);
		if (Lang)
		{
			LgiResources *r = LgiGetResObj();
			if (r)
			{
				int n = 0;
				GLanguage *Cur = LgiGetLanguageId();

				Lang->Empty();
				for (GLanguageId *l = r->GetLanguages(); *l; l++)
				{
					GLanguage *Info = GFindLang(*l);
					if (Info)
					{
						GMenuItem *Item = Lang->AppendItem(Info->Name, IDM_LANG_BASE+n++, true);
						if (Item AND Cur == Info)
						{
							Item->Checked(true);
						}
					}
				}
			}
		}
	}
	else
	{
		printf("AppWnd::SetupUi() Couldn't load menu, missing 'iftp.lr8'?\n");
		LgiExitApp();
	}

	Commands = LgiLoadToolbar(this, "_commands.gif");
	if (Commands)
	{
		Commands->TextLabels(true);
		Commands->Attach(this);
		Commands->Customizable(GetOptions(), "maintools");

		CmdConnect.ToolButton = Commands->AppendButton(LgiLoadString(IDS_CONNECT), IDM_CONNECT, TBT_PUSH);
		CmdDisconnect.ToolButton = Commands->AppendButton(LgiLoadString(IDS_DISCONNECT), IDM_DISCONNECT, TBT_PUSH, false);
		// CmdAbort.ToolButton = Commands->AppendButton(LgiLoadString(IDS_ABORT), IDM_ABORT, TBT_PUSH, false);
		Commands->AppendSeparator();
		
		char *Hidden = LgiLoadString(IDS_SHOWHIDDEN);
		if (Hidden)
		{
			char Str[256];
			strsafecpy(Str, Hidden, sizeof(Str));
			
			// Turn every 2nd space into a non-breaking space
			int n = 0;
			for (char *Sp = Str; *Sp; Sp++)
			{
				if (*Sp == ' ')
				{
					if (n++ % 2 == 0)
					{
						memmove(Sp+1, Sp, strlen(Sp)+1);
						int Len = 2;
						LgiUtf32To8(0xa0, (uint8 *&)Sp, Len);
					}
				}
			}
			
			Commands->AppendButton(Str, IDM_SHOW_HIDDEN, TBT_TOGGLE, true, 3);
		}
		
		SetCtrlValue(IDM_SHOW_HIDDEN, ShowHiddenFiles());
	}

	char *FileName = LgiFindFile("_icons.gif");
	if (FileName)
	{
		GSurface *pDC = LoadDC(FileName);
		if (pDC)
		{
			ImageList = NEW(GImageList(16, 16, pDC));
			#ifdef WIN32
			if (ImageList)
			{
				ImageList->Create(pDC->X(), pDC->Y(), GdcD->GetBits());
			}
			#endif

			DeleteObj(pDC);
		}
		DeleteArray(FileName);
	}

	Status = NEW(GStatusBar);
	if (Status)
	{
		Status->Attach(this);
		StatusInfo[0] = Status->AppendPane("", 2);
		Progress = NEW(FileTransferProgress(this, Status, StatusInfo+1, true));
	}

	LogPanel = NEW(GLogPanel);
	if (LogPanel)
	{
		Pour();
		LogPanel->SetClosedSize((Commands) ? Commands->Y()-1 : 16);
		
		int LogOpen = true;
		GetOptions()->Get(OPT_LogOpen, LogOpen);
		LogPanel->Open(LogOpen);
		
		LogPanel->Attach(this);
	}

	// Worker thread doesn't need to suck too much CPU
	if (Worker)
	{
		Worker->Priority(PRIORITY_LOW);
	}

	// Split the local / remote views..
	Splitter = NEW(GSplitter);
	if (Splitter)
	{
		int Pos = 400;
		Options.Get("split-pos", Pos);

		Splitter->Attach(this);
		Splitter->Value(Pos);

		Local = NEW(GLocalFS(this, ImageList));
		Remote = NEW(GRemoteFS(this, ImageList));

		if (Local)
		{
			Local->Worker = Worker;
			Local->Opposite = Remote;
			Splitter->SetViewA(Local, false);
		}

		if (Remote)
		{
			Remote->Worker = Worker;
			Remote->Opposite = Local;
			Splitter->SetViewB(Remote, false);
		}
	}

	EnableCommands(false);
	UpdateMenus();

	#if defined WIN32
	Pour();
	ShowWindow(Handle(), LgiApp->GetShow());
	UpdateWindow(Handle());
	#else
	Visible(true);
	#endif
}

void AppWnd::UpdateMenus()
{
	if (AccountLst)
	{
		// empty current menu
		while (AccountLst->RemoveItem(0));

		// refill with current accounts
		int i=0;
		for (GTag *p=Accounts.First(); p; p=Accounts.Next(), i++)
		{
			GVariant Name;
			if (p->GetVariant(OPT_AccountName, Name))
			{
				AccountLst->AppendItem(Name.Str(), IDM_ACCOUNT_BASE+i, true);
			}
		}

		if (NOT AccountLst->ItemAt(0))
		{
			char s[256];
			sprintf(s, "<%s>", LgiLoadString(IDS_NONE));
			AccountLst->AppendItem(s, -1, false);
		}
	}
}

void AppWnd::ResetThread()
{
	if (Worker->IsWorking() AND Ftp)
	{
		// close the files and sockets...
		Ftp->Close();
	}

	// terminate...
	if (Worker)
	{
		Worker->Terminate();
		DeleteObj(Worker);
	}

	// setup a new thread...
	Worker = NEW(FtpWorkerThread);

	if (Local)
	{
		Local->Worker = Worker;
	}

	if (Remote)
	{
		Remote->Worker = Worker;
	}
}

bool AppWnd::GetBinaryMode(char *Name)
{
	char *Types = 0;
	
	if (Name)
	{
		if (Options.Get(OPT_AsciiTypes, Types))
		{
			GToken t(Types, ";, ");
			for (int i=0; i<t.Length(); i++)
			{
				if (MatchStr(t[i], Name))
				{
					// Ascii
					return false;
				}
			}
		}
		if (Options.Get(OPT_BinaryTypes, Types))
		{
			GToken t(Types, ";, ");
			for (int i=0; i<t.Length(); i++)
			{
				if (MatchStr(t[i], Name))
				{
					// Binary
					return true;
				}
			}
		}
	}

	// the default
	return BinaryMode;
}

void AppWnd::EnableCommands(bool Enabled)
{
	CmdConnect.Enabled(NOT Enabled);
	CmdHttp.Enabled(NOT Enabled);
	CmdDisconnect.Enabled(Enabled);
	CmdAbort.Enabled(false);

	if (Local)
	{
		Local->EnableCommands(Enabled);
	}

	if (Remote)
	{
		Remote->EnableCommands(Enabled);
	}

	if (Enabled)
	{
		LastCmd = LgiCurrentTime();
		SetPulse(100);
	}
	else
	{
		LastCmd = 0;
		SetPulse();
	}
}

void AppWnd::OnCmd(FtpCmd *Cmd)
{
	if (NOT Cmd->View->IsCmdLocal(Cmd->Cmd))
	{
		LastCmd = LgiCurrentTime();
	}

	if (Ftp AND
		NOT Ftp->IsOpen())
	{
		EnableCommands(false);
	}
}

void AppWnd::IsValid()
{
	/*
	for (FtpAcc *a=Accounts.First(); a; a=Accounts.Next())
	{
		GVariant Name, Host;
		a->GetVariant(OPT_RemoteHost, Host);
		char *h = Host.Str();

		if (a->GetVariant(OPT_AccountName, Name) AND
			Name.Str() AND
			stricmp(Name.Str(), "Test") == 0)
		{
			if (NOT h OR stricmp(h, "ftp.mozilla.org") != 0)
			{
				_asm int 3
			}
		}
	}
	*/
}

void AppWnd::OnUrl(char *Url)
{
	if (strnicmp(Url, "ftp://", 6) == 0)
	{
		// Ftp URL
		FtpAcc Account;
		char *File = 0;
		ConvertUrlToAccount(Url, &Account, &File);
		ConnectTo(&Account);
		
		if (File)
		{
			PostEvent(FTP_SAVE_FILE, 0, (int)NewStr(File));
		}
	}
	else if (strnicmp(Url, "http://", 7) == 0)
	{
		// Http URL
		HttpDlg *Dlg = NEW(HttpDlg(this, Url));
		if (Dlg AND Dlg->DoModal() == IDOK)
		{
			Http = NEW(HttpThread(this, Dlg));
		}
		else
		{
			DeleteObj(Dlg);
		}
	}
}

int AppWnd::OnDrop(char *Format, GVariant *Data, GdcPt2 Pt, int KeyState)
{
	int Status = DROPEFFECT_NONE;

	if (Format AND
		Data AND
		Data->Type == GV_BINARY)
	{
		char *Url = 0;
		char *d = (char*)Data->Value.Binary.Data;

		if (stricmp(Format, "UniformResourceLocator") == 0)
		{
			if ((d[0] == 'f' OR d[0] == 'h') AND d[1] == 0)
			{
				// probably a utf-16 string... naughty, try and convert
				Url = LgiNewUtf16To8((char16*)d);
			}
			else
			{
				Url = NewStr(d);
			}
		}
		else if (stricmp(Format, "text/html") == 0)
		{
			char *Html = LgiNewUtf16To8((char16*)d);
			Xml x;
			if (x.ParseXml(Html))
			{
				for (XmlTag *t=x.Values.First(); t; t=x.Values.Next())
				{
					if (stricmp(t->Name, "a") == 0)
					{
						char *u;
						if (t->Get("href", u))
						{
							Url = NewStr(u);
						}
						break;
					}
				}
			}
			DeleteArray(Html);
		}
		else if (stricmp(Format, LGI_FileDropFormat) == 0)
		{
			GDropFiles Files(*Data);

			for (char *File = Files.First(); File; File = Files.Next())
			{
				if (DirExists(File))
				{
					// Move to the local directory pane to the path of the incomming file
					Local->Cd(File);
					break;
				}
				else if (FileExists(File))
				{
					// Upload list of files...
					char d[300];
					strsafecpy(d, File, sizeof(d));
					LgiTrimDir(d);
					Local->Cd(d);

					// If we are connected upload the file...??
					DirListItem *e = 0;
					char *c = strrchr(File, DIR_CHAR);
					if (c AND
						(e = Local->HasEntry(c + 1)) AND
						Ftp->IsOpen())
					{
						FtpCmd *Cmd = NEW(FtpCmd);
						if (Cmd)
						{
							Cmd->Cmd = TASK_Transfer;
							Cmd->Entry = NEW(IFtpEntry(e->GetEntry()));
							Cmd->View = Local;

							if (Local->Opposite)
							{
								FtpCmd *ListCmd = NEW(FtpCmd);
								if (ListCmd)
								{
									ListCmd->Cmd = TASK_ListElements;
									ListCmd->View = Local->Opposite;
									ListCmd->ForceRemote = NOT Local->IsCmdLocal(TASK_ListElements);
									Cmd->OnSuccess.Insert(ListCmd);
								}
							}

							Local->PostCmd(Cmd);
						}
					}
				}
			}
		}

		if (Url)
		{
			OnUrl(Url);
			Status = DROPEFFECT_COPY;
		}

		DeleteArray(Url);
	}

	return Status;
}

int AppWnd::WillAccept(List<char> &Formats, GdcPt2 Pt, int KeyState)
{
	int Status = DROPEFFECT_NONE;
	for (char *i=Formats.First(); i; )
	{
		if (stricmp(i, "UniformResourceLocator") == 0 OR
			stricmp(i, "text/html") == 0 OR
			stricmp(i, LGI_FileDropFormat) == 0)
		{
			// hello girls.... a URL or file...
			Status = DROPEFFECT_COPY;
			i = Formats.Next();
		}
		else
		{
			// tell the d'n'd code that we don't support this
			Formats.Delete(i);
			DeleteArray(i);
			i = Formats.Current();
		}
	}

	return Status;
}

int AppWnd::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case FTP_SAVE_FILE:
		{
			char *File = (char*)MsgB(Msg);
			char *Folder = (char*)MsgA(Msg);
			if (File)
			{
				if (Folder)
				{
					Local->Cd(Folder, true);
				}
				else
				{
					GFileSelect s;
					s.Name(File);
					s.Parent(this);
					if (s.Save())
					{
						char *Out = s.Name();
						LgiTrimDir(Out);
						Local->Cd(Out);
					}
				}

				FtpCmd *c = NEW(FtpCmd(File));
				if (c)
				{
					c->Cmd = TASK_Transfer;
					c->View = Remote;
					Remote->PostCmd(c);
				}

				DeleteArray(File);
				DeleteArray(Folder);
			}
			break;
		}
		case M_CHANGE:
		{
			return OnNotify((GView*) MsgA(Msg), MsgB(Msg));
		}
		case M_DESCRIBE:
		{
			char *Text = (char*) MsgB(Msg);
			if (Text)
			{
				// SetStatusText(Text, 0);
			}
			break;
		}
		case M_CLOSE:
		{
			int asd=0;
			break;
		}
	}
	
	return GWindow::OnEvent(Msg);
}

void AppWnd::OnDisconnect()
{
	EnableCommands(false);

	if (Remote)
	{
		Remote->SetCtrlName(GFSV_DIR, "");
		Remote->Empty();
	}

	if (StatusInfo[0])
	{
		StatusInfo[0]->Name("");
	}
	
	if (Progress)
	{
		Progress->Value(0);
	}
}

void AppWnd::ConnectTo(FtpAcc *Account)
{
	DeleteObj(Ftp);

	if (Account)
	{
		GVariant SecureFtp;
		Account->GetVariant(OPT_SFtp, SecureFtp);

		if (SecureFtp.CastInt32())
		{
			Ftp = NEW(SFtp(this));
		}
		else
		{
			Ftp = NEW(IFtp(this));
		}

		if (Ftp)
		{
			Ftp->SetMeter(Progress);

			char *Socks5Server = 0;
			char *Socks5UserName = 0;
			char *Socks5Password = 0;

			Options.Get(OPT_Socks5Server, Socks5Server);
			Options.Get(OPT_Socks5UserName, Socks5UserName);
			Options.Get(OPT_Socks5Password, Socks5Password);
			
			GVariant NormList;
			Account->GetVariant(OPT_NormalList, NormList);

			// Ftp protocol class
			Ftp->IsLongList(NOT NormList.Value.Int);

			Connection = 0;
			if (ValidStr(Socks5Server))
			{
				char Buf[256];
				strcpy(Buf, Socks5Server);
				char *Colon = strchr(Buf, ':');
				int Port = 1080;
				if (Colon)
				{
					Port = atoi(Colon+1);
					*Colon = 0;
				}

				Connection = NEW(ISocks5LogConnection(	Buf,
														Port,
														Socks5UserName,
														Socks5Password,
														LogPanel,
														this));
			}
			else
			{
				Connection = NEW(ILogConnection(LogPanel, this));
			}

			char *Log;
			if (Connection AND
				GetOptions()->Get(OPT_LogFile, Log) AND
				ValidStr(Log))
			{
				Connection->SetLogFile(Log);
			}

			GVariant RemoteHost;
			GVariant Port;
			GVariant UserName;
			GVariant Password;
			GVariant RemoteDir;
			GVariant LocalDir;
			GVariant Anonymous;
			GVariant UseActive;
			char CmdStr[256];

			Account->GetVariant(OPT_RemoteHost, RemoteHost);
			Account->GetVariant(OPT_Port, Port);
			Account->GetVariant(OPT_UserName, UserName);
			Account->GetVariant(OPT_Password, Password);
			Account->GetVariant(OPT_Anonymous, Anonymous);
			Account->GetVariant(OPT_LocalDir, LocalDir);
			Account->GetVariant(OPT_RemoteDir, RemoteDir);
			Account->GetVariant(OPT_UseActive, UseActive);

			// Set local dir
			if (LocalDir.Str() AND Local)
			{
				FtpCmd *Cmd = NEW(FtpCmd(LocalDir.Str()));
				if (Cmd)
				{
					Cmd->Cmd = TASK_Cd;
					Cmd->View = Local;
					Local->PostCmd(Cmd);
				}
			}

			// Loging to remote server
			char *Rh = TrimStr(RemoteHost.Str());
			LogPanel->Print(FTP_COL_INFO, LgiLoadString(IDS_CONNECT_TO), RemoteHost.Str());
			if (Anonymous.IsInt() AND Anonymous.Value.Int)
			{
				sprintf(CmdStr,
						"%s,%i,%s,%s,%s,%i",
						Rh,
						Port.Value.Int,
						"anonymous",
						"user@localhost.com",
						(ValidStr(RemoteDir.Str()))?RemoteDir.Str():" ",
						UseActive.Value.Int);
			}
			else
			{
				sprintf(CmdStr,
						"%s,%i,%s,%s,%s,%i",
						Rh,
						Port.Value.Int,
						UserName.Str(),
						Password.Str(),
						(ValidStr(RemoteDir.Str()))?RemoteDir.Str():" ",
						UseActive.Value.Int);
			}
			DeleteArray(Rh);

			// Enabled abort command
			Connecting = true;

			// Post login command
			FtpCmd *Cmd = NEW(FtpCmd(CmdStr));
			if (Cmd)
			{
				Cmd->Cmd = TASK_Login;
				Cmd->View = Remote;
				Cmd->Connection = Connection;
				Local->PostCmd(Cmd);
			}
		}
	}
}

#define MenuCommand(fName)				\
	case IDM_##fName:					\
	{									\
		GView *f = LgiApp->GetFocus();	\
		if (f) f->PostEvent(M_##fName);	\
		break;							\
	}

int AppWnd::OnCommand(int Cmd, int Event, OsView Handle)
{
	if (Cmd >= IDM_ACCOUNT_BASE AND Cmd < Accounts.GetItems()+IDM_ACCOUNT_BASE)
	{
		ConnectTo(Accounts.ItemAt(Cmd - IDM_ACCOUNT_BASE));
	}

	switch (Cmd)
	{
		// File menu
		case IDM_OPTIONS:
		{
			OptionsDlg Dlg(this, &Options);
			break;
		}
		
		// Type menu
		case IDM_ASCII:
		{
			Modes[0]->Checked(true);
			Modes[1]->Checked(false);
			BinaryMode = false;
			break;
		}
		case IDM_BINARY:
		{
			Modes[0]->Checked(false);
			Modes[1]->Checked(true);
			BinaryMode = true;
			break;
		}
		case IDM_HTTP:
		{
			HttpDlg *Dlg = NEW(HttpDlg(this, "http://"));
			if (Dlg AND Dlg->DoModal() == IDOK)
			{
				Http = NEW(HttpThread(this, Dlg));
			}
			else
			{
				DeleteObj(Dlg);
			}
			break;
		}
		case IDM_CONNECT:
		{
			{
				ConnectDlg Dlg(this, &Accounts);
				ConnectTo(Dlg.Account);
			}
			UpdateMenus();
			break;
		}
		case IDM_DISCONNECT:
		{
			if (Ftp)
			{
				ResetThread();
				Remote->Empty();
				Ftp->Close();
				EnableCommands(false);
			}
			break;
		}

		case IDM_ABORT:
		{
			if (Connecting)
			{
				ResetThread();

				Remote->Empty();
				Ftp->Close();
				EnableCommands(false);
				Connecting = false;

				LogPanel->Print(FTP_COL_INFO, LgiLoadString(IDS_CONNECT_CANCEL));
			}
			else if (Progress)
			{
				Progress->Cancel(true);
			}
			break;
		}
		case IDM_EXIT:
		{
			LgiCloseApp();
			break;
		}
		case IDM_SCHEDULE:
		{
			NEW(ScheduleWnd(this));
			break;
		}
		case IDM_HELP:
		{
			char *File = "index.html";
			char *Path = LgiFindFile(File);
			if (Path AND
				LgiExecute(Path, NULL, NULL))
			{
			}
			else
			{
				LgiMsg(this, LgiLoadString(IDS_ERR_NO_HELP_FILE), AppName, MB_OK, File);
			}
			DeleteArray(Path);
			break;
		}
		case IDM_ABOUT:
		{
			GAbout Dlg(this, AppName, VER, LgiLoadString(IDS_ABOUT), "_about.gif");
			break;
		}
		default:
		{
			if (Cmd >= IDM_LANG_BASE AND Cmd < IDM_LANG_BASE + 100)
			{
				int Index = Cmd - IDM_LANG_BASE;
				
				LgiResources *r = LgiGetResObj();
				if (r)
				{
					GLanguageId Lang = r->GetLanguages()[Index];
					if (Lang)
					{
						Options.Set(OPT_Language, Lang);
						if (LgiMsg(this, LgiLoadString(IDS_RESTART), AppName, MB_YESNO) == IDYES)
						{
							Restart = true;
							LgiCloseApp();
						}
					}
				}
			}
			break;
		}
	}

	return 0;
}

int AppWnd::OnNotify(GView *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		case IDM_SHOW_HIDDEN:
		{
			int Sh = GetCtrlValue(IDM_SHOW_HIDDEN);

			Options.Set(OPT_ShowHidden, Sh);
			
			Local->ListElements();

			if (Ftp)
			{
				Ftp->IsShowHidden(Sh);
			}
			break;
		}
	}
	
	return 0;
}

void AppWnd::WriteProperties(GFile &F, ObjProperties *Props, char *KeyName)
{
	if (Props AND KeyName)
	{
		char *Str = NEW(char[16<<10]);
		if (Str)
		{
			*Str = 0;
			sprintf(Str, "<%s", KeyName);
			for (bool b = Props->FirstKey(); b; b = Props->NextKey())
			{
				switch (Props->KeyType())
				{
					case OBJ_INT:
					{
						int n;
						if (Props->Get(Props->KeyName(), n))
						{
							sprintf(Str+strlen(Str), "\r\n\t%s=%i", Props->KeyName(), n);
						}
						break;
					}
					case OBJ_STRING:
					{
						char *s;
						if (Props->Get(Props->KeyName(), s) AND ValidStr(s))
						{
							sprintf(Str+strlen(Str), "\r\n\t%s='%s'", Props->KeyName(), s);
						}
						break;
					}
				}
			}

			sprintf(Str+strlen(Str), ">\r\n");
			F.Write(Str, strlen(Str));
			DeleteArray(Str);
		}
	}
}

void AppWnd::ReadProperties(XmlTag *t, ObjProperties *a)
{
	if (t AND a)
	{
		for (XmlValue *v = t->Values.First(); v; v = t->Values.Next())
		{
			bool IsNumber = true;
			for (char *s = v->Value; *s AND IsNumber; s++)
			{
				IsNumber &= (*s >= '0' AND *s <= '9');
			}

			if (IsNumber)
			{
				a->Set(v->Name, atoi(v->Value));
			}
			else
			{
				a->Set(v->Name, v->Value);
			}
		}
	}
}

void AppWnd::Serialize(bool Write)
{
	char *AccountsKeyName = "account";
	char *OptionsKeyName = "options";

	if (NOT OptionsFile)
	{
		char File[256];
		if (LgiApp->GetOption("o", File))
		{
			if (FileExists(File))
			{
				OptionsFile = NewStr(File);
			}
			else
			{
				char *Temp = LgiFindFile(File);
				if (Temp)
				{
					OptionsFile = Temp;
				}
				else
				{
					OptionsFile = NewStr(File);
				}
			}
		}

		if (NOT OptionsFile)
		{
			OptionsFile = LgiFindFile(OptionsFileName);
		}

		if (NOT OptionsFile)
		{
			char Temp[256];
			if (LgiGetExePath(Temp, sizeof(Temp)))
			{
				#ifdef _DEBUG
				LgiTrimDir(Temp);
				#endif
				strcat(Temp, DIR_STR);
				strcat(Temp, OptionsFileName);
				OptionsFile = NewStr(Temp);
			}
		}
	}

	GFile F;
	if (OptionsFile AND
		F.Open(OptionsFile, (Write)?O_WRITE:O_READ))
	{
		if (Write)
		{
			F.SetSize(0);

			// Save window position
			#ifdef WIN32
			if (Handle() AND NOT IsIconic(Handle()) AND NOT IsZoomed(Handle()))
			#endif
			{
				GRect r = GetPos();
				Options.Set("x", r.x1);
				Options.Set("y", r.y1);
				Options.Set("cx", r.X()-1);
				Options.Set("cy", r.Y()-1);
			}

			WriteProperties(F, &Options, OptionsKeyName);
			for (FtpAcc *a = Accounts.First(); a; a = Accounts.Next())
			{
				a->Write(F);
				// WriteProperties(F, a, AccountsKeyName);
			}
		}
		else
		{
			Xml X;
			if (X.ParseXmlFile(F) > 0)
			{
				for (XmlTag *t = X.Values.First(); t; t = X.Values.Next())
				{
					if (t->Name)
					{
						if (stricmp(t->Name, AccountsKeyName) == 0)
						{
							FtpAcc *a = NEW(FtpAcc);
							if (a)
							{
								// ReadProperties(t, a);
								if (a->Read(t))
								{
									Accounts.Insert(a);
								}
								else
								{
									DeleteObj(a);
								}
							}
						}
						else if (stricmp(t->Name, OptionsKeyName) == 0)
						{
							ReadProperties(t, &Options);
						}
					}
				}
			}

			GRect r;
			if (Options.Get("x", r.x1) AND
				Options.Get("y", r.y1) AND
				Options.Get("cx", r.x2) AND
				Options.Get("cy", r.y2))
			{
				r.x2 += r.x1;
				r.y2 += r.y1;
				SetPos(r);
			}
		}
	}
}

void AppWnd::SetTransfer(bool On)
{
	if (Progress)
	{
		Progress->Cancel(false);
	}

	CmdAbort.Enabled(On);
}

void AppWnd::OnScheduleChanged(char *Xml)
{
	if (NOT d->Schedule)
	{
		d->Schedule = NEW(GXmlTag);
	}
	if (d->Schedule)
	{
		GFile f;
		if (f.Open(Xml, O_READ))
		{
			GXmlTree t;
			t.Read(d->Schedule, &f, 0);
			SetPulse(1000);
		}
	}
}

//////////////////////////////////////////////////////////////////////
void FtpList::OnColumnClick(int Col, GMouse &m)
{
	if (View)
	{
		View->OnColumnClick(Col, m);
	}
}
