#include <ctype.h>

#include "Ftp.h"
#include "GButton.h"
#include "GCombo.h"
#include "GEdit.h"

GLocalFS::GLocalFS(AppWnd *Wnd, GImageList *ImageList) :
	GFileSystemView(Wnd, ImageList, true)
{
	Children.Insert(Dir = NEW(GButton(IDC_DIR, 0, 0, 70, 20, LgiLoadString(IDS_EXPLORE))));
	Children.Insert(Mount = NEW(GCombo(IDC_MOUNT, 0, 0, 50, 20, "")));
	if (Mount)
	{
		Mount->Value(-1);

		GVolume *v = FileDev->GetRootVolume();
		if (v)
		{
			Mount->Insert(LgiLoadString(IDS_DESKTOP));

			for (GVolume *d=v->First(); d; d=v->Next())
			{
				if (d->Type() != VT_RAMDISK)
				{
					Mount->Insert(d->Path());
				}
			}
		}
	}
}

void GLocalFS::EnableCommands(bool Enabled)
{
	CmdTransfer.Enabled(Enabled);
}

bool GLocalFS::IsCmdLocal(int Cmd)
{
	return (Cmd != TASK_Transfer);
}

bool GLocalFS::ListElements()
{
	bool Status = false;

	if (Listing)
	{
		bool DiffFolder = NOT CurFolder OR stricmp(CurFolder, Edit->Name()) != 0;
		DeleteArray(CurFolder);
		CurFolder = NewStr(Edit->Name());

		List<GListItem> Temp;
		GHashTable Hash;
		if (DiffFolder)
		{
			// Clear all previous items
			Empty();
			Lst->Empty();

			bool IsRoot = false;
			#ifdef WIN32
			IsRoot = strlen(CurFolder) <= 3 AND CurFolder[0] AND CurFolder[1] == ':';
			#else
			IsRoot = stricmp(CurFolder, "/") == 0;
			#endif

			if (NOT IsRoot)
			{
				IFtpEntry *DotDot = NEW(IFtpEntry);
				if (DotDot)
				{
					DotDot->Attributes = IFTP_DIR;
					DotDot->Name = NewStr("..");
					DirListItem *Item = NEW(DirListItem(DotDot, this));
					if (Item)
					{
						Temp.Insert(Item);
					}
				}
			}
		}
		else
		{
			// Add all old items to the hash table
			for (DirListItem *i=dynamic_cast<DirListItem*>(Lst->First());
							  i;
							  i=dynamic_cast<DirListItem*>(Lst->Next()))
			{
				Hash.Add(i->Entry->Name, i);
			}
		}

		GDirectory *Dir = FileDev->GetDir();
		if (Dir)
		{
			for (int Status = Dir->First(Edit->Name()); Status; Status = Dir->Next())
			{
				bool Hidden = Dir->IsHidden();
				if (NOT Wnd->ShowHiddenFiles() AND Hidden)
				{
					// Do nothing...
				}
				else
				{
					// Create entry
					IFtpEntry *e = NEW(IFtpEntry);
					if (e)
					{
						e->Name = NewStr(Dir->GetName());
						e->Size = Dir->GetSize();

						if (Dir->IsDir()) e->Attributes |= IFTP_DIR;

						#ifdef WIN32
						if (Dir->GetAttributes() & FA_READONLY) e->Attributes |= IFTP_READ;
						if (Dir->GetAttributes() & FA_ARCHIVE) e->Attributes |= IFTP_ARCHIVE;
						if (Dir->GetAttributes() & FA_HIDDEN) e->Attributes |= IFTP_HIDDEN;
						if (Dir->GetAttributes() & FA_SYSTEM) e->Attributes |= IFTP_SYSTEM;
						#else
						if (Dir->GetAttributes() & S_IRUSR) e->Attributes |= IFTP_READ;
						if (Dir->GetAttributes() & S_IWUSR) e->Attributes |= IFTP_WRITE;
						if (Dir->GetAttributes() & S_IXUSR) e->Attributes |= IFTP_EXECUTE;

						if (Dir->GetAttributes() & S_IRGRP) e->Attributes |= IFTP_GRP_READ;
						if (Dir->GetAttributes() & S_IWGRP) e->Attributes |= IFTP_GRP_WRITE;
						if (Dir->GetAttributes() & S_IXGRP) e->Attributes |= IFTP_GRP_EXECUTE;

						if (Dir->GetAttributes() & S_IROTH) e->Attributes |= IFTP_GLOB_READ;
						if (Dir->GetAttributes() & S_IWOTH) e->Attributes |= IFTP_GLOB_WRITE;
						if (Dir->GetAttributes() & S_IXOTH) e->Attributes |= IFTP_GLOB_EXECUTE;
						#endif

						if (Hidden)
						{
							e->Attributes |= IFTP_HIDDEN;
						}

						uint64 GDateTime = Dir->GetLastWriteTime();
						char Str[256];
						Dir->ConvertToDate(Str, GDateTime);
						strcat(Str, " ");
						Dir->ConvertToTime(Str+strlen(Str), GDateTime);
						e->Date.Set(Str);

						if (Dir->Path(Str, sizeof(Str)))
						{
							e->Path = NewStr(Str);
						}

						if (DiffFolder)
						{
							// Just add it... we're scanning a different folder...
							Insert(e);
							Temp.Insert(NEW(DirListItem(e, this)), (e->IsDir()) ? 0 : -1);
						}
						else
						{
							// Scanning the same folder...

							// Is it already there?
							DirListItem *Old = (DirListItem*) Hash.Find(e->Name);
							if (Old)
							{
								// Update the old...
								Old->Entry->Size = e->Size;
								Old->Entry->Date = e->Date;
								Old->Entry->Attributes = e->Attributes;
								
								// Clear the hash table entry...
								// Because when we're done the hash table holds all the entries
								// that have been deleted.
								Hash.Delete(Old->Entry->Name);

								// Clean up
								DeleteObj(e);
								Old->Update();
							}
							else
							{
								// New entry
								DirListItem *New = NEW(DirListItem(e, this));
								if (New)
								{
									Insert(e);
									Lst->Insert(New, (e->IsDir()) ? 0 : -1);
								}
							}
						}
					}
				}
			}

			Status = true;
			DeleteObj(Dir);
		}

		if (Lst)
		{
			if (DiffFolder)
			{
				Temp.Sort(FtpItemCompare, (int)this);
				Lst->Insert(Temp);
			}
			else
			{
				// Clear deleted items
				for (DirListItem *Del = (DirListItem*)Hash.First(); Del; Del = (DirListItem*)Hash.Next())
				{
					if (stricmp(Del->Entry->Name, ".."))
					{
						Lst->Remove(Del);
						DeleteObj(Del);
					}
				}

				// Sort the remaining items
				Lst->Sort(FtpItemCompare, (int)this);
			}

			int Resize = 0;
			if (Wnd->GetOptions()->Get(OPT_ResizeToContent, Resize) AND
				Resize)
			{
				Lst->ResizeColumnsToContent();
			}
		}
	}
	
	return Status;
}

bool GLocalFS::Cd(char *To, bool RefeashList)
{
	bool Status = false;

	if (NOT Edit) return false;

	char Dir[300];
	if (To)
	{
		strsafecpy(Dir, To, sizeof(Dir));
	}
	else
	{
		strsafecpy(Dir, Edit->Name(), sizeof(Dir));
		if (strlen(Dir) > 2)
		{
			LgiTrimDir(Dir);
		}
	}

	#ifdef WIN32
	if (strlen(Dir) == 2 AND
		Dir[1] == ':')
	{
		strcat(Dir, "\\");
	}
	#endif

	if (Dir AND DirExists(Dir))
	{
		Edit->Name(Dir);
		if (RefeashList)
		{
			ListElements();
		}
		Status = true;
	}


	return Status;
}

bool GLocalFS::CreateDir(char *Dir)
{
	if (Dir)
	{
		char p[300];
		if (LgiMakePath(p, sizeof(p), CurFolder, Dir))
		{		
			if (FileDev->CreateDirectory(p))
			{
				ListElements();
				return true;
			}
		}
	}

	return false;
}

bool GLocalFS::Delete(IFtpEntry *e)
{
	if (e)
	{
		if (e->IsDir())
		{
			return FileDev->RemoveDirectory(e->Path);
		}
		else
		{
			return FileDev->DeleteFile(e->Path);
		}
	}

	return false;
}

bool GLocalFS::SetPerms(IFtpEntry *e, int Perms)
{
	bool Status = false;
	if (e)
	{
		char Path[256];
		GetDir(Path, e->Name);

		#if defined WIN32
		SetFileAttributes(	Path,
							((Perms & IFTP_READ) ? FILE_ATTRIBUTE_READONLY : 0) |
							((Perms & IFTP_ARCHIVE) ? FILE_ATTRIBUTE_ARCHIVE : 0) |
							((Perms & IFTP_HIDDEN) ? FILE_ATTRIBUTE_HIDDEN : 0) |
							((Perms & IFTP_SYSTEM) ? FILE_ATTRIBUTE_SYSTEM : 0));
		#elif defined BEOS
		BEntry Entry(Path);
		int p = 0;

		if (Perms & IFTP_READ) p |= S_IRUSR;
		if (Perms & IFTP_WRITE) p |= S_IWUSR;
		if (Perms & IFTP_EXECUTE) p |= S_IXUSR;

		if (Perms & IFTP_GRP_READ) p |= S_IRGRP;
		if (Perms & IFTP_GRP_WRITE) p |= S_IWGRP;
		if (Perms & IFTP_GRP_EXECUTE) p |= S_IXGRP;

		if (Perms & IFTP_GLOB_READ) p |= S_IROTH;
		if (Perms & IFTP_GLOB_WRITE) p |= S_IWOTH;
		if (Perms & IFTP_GLOB_EXECUTE) p |= S_IXOTH;
		#endif
	}

	return Status;
}

bool GLocalFS::ViewTextFile(char *File)
{
	bool Status = false;
	char App[256];
	if (File AND
		LgiGetAppForMimeType("text/plain", App, sizeof(App)))
	{
		bool ArgInserted = false;
		char Path[256];
		GetDir(Path, File);

		for (char *a=strchr(App, '%'); a AND *a; a=strchr(a, '%'))
		{
			if (isdigit(a[1]) OR
				a[1] == 'L' OR
				a[1] == 'f')
			{
				// argument placeholder
				if (NOT ArgInserted)
				{
					int Len = strlen(Path);
					a += 2;
					memmove(a + Len - 2, a, strlen(a) + 1);
					memcpy(a - 2, Path, Len);
					ArgInserted = true;
				}
				break;
			}
			else
			{
				// environment variable
				char *End = strchr(a+1, '%');
				if (End)
				{
					*End++ = 0;
					if (stricmp(a+1, "systemroot") == 0)
					{
						char Temp[256];
						#ifdef WIN32
						GetWindowsDirectory(Temp, sizeof(Temp));
						#else
						strcpy(Temp, "/tmp");
						#endif
						int Len = strlen(Temp);

						if (Temp[strlen(Temp)-1] == DIR_CHAR) Temp[strlen(Temp)-1] = 0;
						strcat(Temp, End);
						strcpy(App, Temp);

						a = App + Len;
					}
				}
				else
				{
					a ++;
				}
			} 
		}

		if (NOT ArgInserted)
		{
			strcat(App, " ");
			strcat(App, Path);
		}

		// find end of program
		#ifdef WIN32
		char *Arg = stristr(App, ".exe");
		#else
		char *Arg = strchr(App, ' ');
		#endif
		if (Arg)
		{
			#ifdef WIN32
			Arg += 4;
			#endif
			if (strchr("\'\"", *Arg))
			{
				Arg++;
			}

			// isolate the arguments...
			*Arg++ = 0;
			Status = LgiExecute(App, Arg, ".");
		}
	}

	return Status;
}

bool GLocalFS::ExecuteFile(char *File)
{
	bool Status = false;
	if (File)
	{
		char Path[256];
		GetDir(Path, File);
		Status = LgiExecute(Path, "", ".");
	}

	return Status;
}

bool GLocalFS::TransferFile(FtpCmd *c)
{
	// Upload file
	bool Status = false;

	if (c AND Wnd->GetFtp())
	{
		char Path[256];
		if (c->Entry)
		{
			if (c->Entry->IsDir())
			{
				return TransferDir(c->Entry, c->Str);
			}
			else
			{
				GetDir(Path, c->Entry->Name);
			}
		}
		else
		{
			strcpy(Path, c->Arg);
		}

		char *File = strrchr(Path, DIR_CHAR);
		if (File)
		{
			Wnd->SetTransfer(true);
			Status = Wnd->GetFtp()->UploadFile(Path, File + 1, Wnd->GetBinaryMode(File + 1));
			Wnd->SetTransfer(false);
		}
	}

	return Status;
}

bool GLocalFS::ProcessClick(IFtpEntry *e, GMouse *m)
{
	bool Status = true;

	if (e)
	{
		if (m->Left() AND m->Double())
		{
			if (e->IsDir())
			{
				// dir
				char Path[300];
				LgiMakePath(Path, sizeof(Path), Edit->Name(), e->Name);
				Status = Cd(Path);
			}
			else
			{
				// file
				FtpCmd *Cmd = NEW(FtpCmd);
				if (Cmd)
				{
					Cmd->Cmd = TASK_Transfer;
					Cmd->Entry = NEW(IFtpEntry(e));
					Cmd->View = this;

					if (Opposite)
					{
						FtpCmd *ListCmd = NEW(FtpCmd);
						if (ListCmd)
						{
							ListCmd->Cmd = TASK_ListElements;
							ListCmd->View = Opposite;
							ListCmd->ForceRemote = NOT IsCmdLocal(TASK_ListElements);
							Cmd->OnSuccess.Insert(ListCmd);
						}
					}

					PostCmd(Cmd);
				}
			}
		}
	}

	return Status;
}

bool GLocalFS::RenameFile(char *From, char *To)
{
	if (From AND To)
	{
		char f[300], t[300];
		if (LgiMakePath(f, sizeof(f), CurFolder, From) AND
			LgiMakePath(t, sizeof(t), CurFolder, To))
		{
			if (FileDev->MoveFile(f, t))
			{
				ListElements();
				return true;
			}
		}
	}

	return false;
}

