#include "Ftp.h"
#include "GEdit.h"

GRemoteFS::GRemoteFS(AppWnd *Wnd, GImageList *ImageList) :
	GFileSystemView(Wnd, ImageList, false)
{
}

void GRemoteFS::EnableCommands(bool Enabled)
{
	CmdCreateDir.Enabled(Enabled);
	CmdDelete.Enabled(Enabled);
	CmdRenameFile.Enabled(Enabled);
	CmdSetPerms.Enabled(Enabled);
	CmdExecute.Enabled(false);
	CmdViewText.Enabled(false);
	CmdTransfer.Enabled(Enabled);
	CmdRefresh.Enabled(Enabled);
	CmdCdup.Enabled(Enabled);
}

bool GRemoteFS::IsCmdLocal(int Cmd)
{
	return Cmd == TASK_ProcessClick;
}

void GRemoteFS::OnItemSelect(GListItem *Item)
{
	if (Wnd->GetFtp() AND Wnd->GetFtp()->IsOpen())
	{
		
	}
}

bool GRemoteFS::ListElements()
{
	bool Status = false;

	if (Listing AND Edit AND Lst AND Wnd->GetFtp())
	{
		// Get servers working directory...
		char DirName[256];
		if (Wnd->GetFtp()->GetDir(DirName))
		{
			// Store the name in the UI
			Edit->Name(DirName);

			// Check whether we've moved location on the server.
			bool DiffFolder = NOT CurFolder OR stricmp(CurFolder, DirName) != 0;
			DeleteArray(CurFolder);
			CurFolder = NewStr(DirName);

			List<GListItem> Temp;
			List<IFtpEntry> Listing;
			GHashTable Hash;
			if (DiffFolder)
			{
				// Clear all previous items
				Empty();
				Lst->Empty();

				bool IsRoot = stricmp(CurFolder, "/") == 0;
				if (NOT IsRoot)
				{
					IFtpEntry *DotDot = NEW(IFtpEntry);
					if (DotDot)
					{
						DotDot->Attributes = IFTP_DIR;
						DotDot->Name = NewStr("..");
						DirListItem *Item = NEW(DirListItem(DotDot, this));
						if (Item)
						{
							Temp.Insert(Item);
						}
					}
				}
			}
			else
			{
				// Add all old items to the hash table
				for (DirListItem *i=dynamic_cast<DirListItem*>(Lst->First());
								  i;
								  i=dynamic_cast<DirListItem*>(Lst->Next()))
				{
					Hash.Add(i->Entry->Name, i);
				}
			}

			// Get the listing off the server..
			if (Status = Wnd->GetFtp()->ListDir(&Listing))
			{
				// Process each item
				for (IFtpEntry *e=Listing.First(); e; e=Listing.Next())
				{
					DeleteArray(e->Path);
					char Path[300];
					if (LgiMakePath(Path, sizeof(Path), CurFolder, e->Name))
					{
						e->Path = NewStr(Path);
					}
					
					if (Wnd->ShowHiddenFiles() OR NOT e->IsHidden())
					{
						if (DiffFolder)
						{
							// Just add it... we're scanning a different folder...
							Insert(e);
							Temp.Insert(NEW(DirListItem(e, this)), (e->IsDir()) ? 0 : -1);
						}
						else
						{
							// Scanning the same folder...

							// Is it already there?
							DirListItem *Old = (DirListItem*) Hash.Find(e->Name);
							if (Old)
							{
								// Update the old...
								Old->Entry->Size = e->Size;
								Old->Entry->Date = e->Date;
								Old->Entry->Attributes = e->Attributes;
								
								// Clear the hash table entry...
								// Because when we're done the hash table holds all the entries
								// that have been deleted.
								Hash.Delete(Old->Entry->Name);

								// Clean up
								DeleteObj(e);
								Old->Update();
							}
							else
							{
								// New entry
								DirListItem *New = NEW(DirListItem(e, this));
								if (New)
								{
									Insert(e);
									Lst->Insert(New, (e->IsDir()) ? 0 : -1);
								}
							}
						}
					}
				}

				if (DiffFolder)
				{
					Temp.Sort(FtpItemCompare, (int)this);
					Lst->Insert(Temp);
				}
				else
				{
					// Clear deleted items
					for (DirListItem *Del = (DirListItem*)Hash.First(); Del; Del = (DirListItem*)Hash.Next())
					{
						if (stricmp(Del->Entry->Name, "..") != 0)
						{
							Lst->Remove(Del);
							DeleteObj(Del);
						}
					}

					// Sort the remaining items
					Lst->Sort(FtpItemCompare, (int)this);
				}

				int Resize = 0;
				if (Wnd->GetOptions()->Get(OPT_ResizeToContent, Resize) AND
					Resize)
				{
					Lst->ResizeColumnsToContent();
				}
			}
		}
	}

	return Status;
}

bool GRemoteFS::Cd(char *Dir, bool RefeashList)
{
	bool Status = false;
	if (Wnd->GetFtp())
	{
		if (Dir)
		{
			char *Slash;
			while (Slash = strchr(Dir, '\\'))
			{
				*Slash = '/';
			}
		}

		if (ValidStr(Dir))
		{
			Status = Wnd->GetFtp()->SetDir(Dir);
		}
		else
		{
			Status = Wnd->GetFtp()->UpDir();
		}

		if (Status AND RefeashList)
		{
			ListElements();
		}
	}

	return Status;
}

bool GRemoteFS::CreateDir(char *Dir)
{
	if (Wnd->GetFtp() AND Dir)
	{
		if (Wnd->GetFtp()->CreateDir(Dir))
		{
			ListElements();
			return true;
		}
	}

	return false;
}

bool GRemoteFS::Delete(IFtpEntry *e)
{
	if (Wnd->GetFtp() AND e)
	{
		if (e->IsDir())
		{
			return Wnd->GetFtp()->DeleteDir(e->Name);
		}
		else
		{
			return Wnd->GetFtp()->DeleteFile(e->Name);
		}
	}

	return false;
}

bool GRemoteFS::SetPerms(IFtpEntry *e, int Perms)
{
	bool Status = false;

	if (Wnd->GetFtp() AND e AND e->Name)
	{
		Wnd->GetFtp()->SetPerms(e->Name, Perms);
	}

	return Status;
}

bool GRemoteFS::TransferFile(FtpCmd *c)
{
	bool Status = false;

	// download file
	if (c AND Opposite)
	{
		char Path[256];
		if (c->Entry)
		{
			if (NOT c->Int AND c->Entry->IsDir())
			{
				return TransferDir(c->Entry, c->Str);
			}
			else
			{
				Opposite->GetDir(Path, c->Entry->Name);
			}
		}
		else
		{
			Opposite->GetDir(Path, c->Arg);

			for (IFtpEntry *e = First(); e; e = Next())
			{
				if (strcmp(e->Name, c->Arg) == 0)
				{
					c->Entry = NEW(IFtpEntry(e));
					break;
				}
			}
		}

		char *File = strrchr(Path, DIR_CHAR);
		if (File AND c->Entry)
		{
			if (FileExists(Path))
			{
				int LocalSize = FileSize(Path);
				bool Resumable = c->Entry->Size > LocalSize;
				
				int Result = IDC_OVERWRITE;
				if (NOT c->Int2)
				{
					ResumeDlg Dlg(this, c->Entry, Resumable);
					Result = Dlg.DoModal();
				}

				if (Result == IDC_OVERWRITE)
				{
					FileDev->DeleteFile(Path);
				}
				else if (Result == IDC_RESUME)
				{
					if (NOT Wnd->GetFtp()->ResumeAt(LocalSize))
					{
						if (LgiMsg(this, LgiLoadString(IDS_RESUME_ERR), AppName, MB_YESNO) == IDNO)
						{
							return false;
						}
					}
				}
				else
				{
					return false;
				}
			}

			Wnd->SetTransfer(true);			
			Status = Wnd->GetFtp()->DownloadFile(Path, c->Entry, Wnd->GetBinaryMode(c->Entry->Name));
			Wnd->SetTransfer(false);
		}
	}

	return Status;
}

class LinkDlg : public GDialog
{
public:
	int Dir;

	LinkDlg(GView *p, int d)
	{
		Dir = d;
		SetParent(p);
		if (LoadFromResource(IDD_LINK))
		{
			MoveToCenter();
			SetCtrlValue(IDC_TYPE, Dir);
		}
	}

	int OnNotify(GView *c, int f)
	{
		switch (c->GetId())
		{
			case IDOK:
			{
				Dir = GetCtrlValue(IDC_TYPE);
			}
			case IDCANCEL:
			{
				EndModal(c->GetId() == IDOK);
				break;
			}
		}
		return 0;
	}
};

bool GRemoteFS::ProcessClick(IFtpEntry *e, GMouse *m)
{
	bool Status = true;

	if (e AND
		e->Name AND
		Opposite AND
		Wnd->GetFtp())
	{
		if (m->Left())
		{
			if (m->Double())
			{
				bool IsDir = e->IsDir();

				if (e->Attributes & IFTP_SYM_LINK)
				{
					// Make a guess
					IsDir = NOT strchr(e->Name, '.');

					int Confirm = 0;
					if (Wnd->GetOptions()->Get(OPT_ConfirmLinks, Confirm) AND
						Confirm)
					{
						LinkDlg d(this, IsDir);
						if (d.DoModal())
						{
							IsDir = d.Dir;
						}
						else
						{
							return false;
						}
					}
				}

				// double click
				if (IsDir)
				{
					// dir

					// change dir to e->Name
					FtpCmd *Cmd = NEW(FtpCmd(e->Name));
					if (Cmd)
					{
						Cmd->Cmd = TASK_Cd;
						Cmd->Entry = NEW(IFtpEntry(e));
						Cmd->View = this;
						PostCmd(Cmd);
					}
				}
				else
				{
					// file
					FtpCmd *Cmd = NEW(FtpCmd);
					if (Cmd)
					{
						Cmd->Cmd = TASK_Transfer;
						Cmd->Entry = e;
						Cmd->View = this;

						// re-list contents after
						FtpCmd *ListCmd = NEW(FtpCmd);
						if (ListCmd)
						{
							ListCmd->Cmd = TASK_ListElements;
							ListCmd->View = Opposite;
							ListCmd->ForceRemote = Opposite->IsCmdLocal(TASK_ListElements);
							Cmd->OnSuccess.Insert(ListCmd);
						}

						PostCmd(Cmd);
					}
				}
			}
			else
			{
				// single click
			}
		}
	}

	return Status;
}

bool GRemoteFS::RenameFile(char *From, char *To)
{
	if (Wnd->GetFtp() AND From AND To)
	{
		if (Wnd->GetFtp()->RenameFile(From, To))
		{
			ListElements();
		}
	}
	return false;
}

