#include "Lgi.h"
#include "Ftp.h"
#include "IHttp.h"

///////////////////////////////////////////////////////////////////////////
class HttpLogConnection : public ILogConnection
{
	char Last[7];

public:
	HttpLogConnection(GLog *log, AppWnd *wnd) :
		ILogConnection(log, wnd)
	{
		ZeroObj(Last);
	}

	void OnRead(char *Data, int Len)
	{
		if (Data)
		{
			memcpy(Last+3, Data, min(3, Len));
			char *EndOfRequest = strnstr(Data, "\r\n\r\n", Len);

			if (strnstr(Last, "\r\n\r\n", Len))
			{
				// EOR on boundry
				SetParameter(GSocket_Log, false);
			}
			
			if (EndOfRequest)
			{
				// EOR in mid block
				LogRead(Data, (int)EndOfRequest-(int)Data);
				SetParameter(GSocket_Log, false);
				ZeroObj(Last);
			}
			else
			{
				// not EOR
				LogRead(Data, Len);
				memcpy(Last, Data+Len-3, 3);
			}
		}
	}
};

///////////////////////////////////////////////////////////////////////////
HttpDlg::HttpDlg(AppWnd *parent, char *Default)
{
	SetParent(Parent = parent);
	Url = 0;
	File = 0;
	Proxy = 0;
	UserName = 0;
	Password = 0;
	StartFromByte = 0;
	NoCache = 0;

	if (LoadFromResource(IDD_HTTP))
	{
		if (Default)
		{
			SetCtrlName(IDC_URL, Default);
		}

		SetCtrlValue(IDC_ROLLBACK_K, 2);
		SetResumable(false);

		MoveToCenter();

		int i;
		if (Parent->GetOptions()->Get(OPT_HttpProxyOn, i))
		{
			SetCtrlValue(IDC_USE_PROXY, i);
		}
		char *s;
		if (Parent->GetOptions()->Get(OPT_HttpProxy, s))
		{
			SetCtrlName(IDC_PROXY_SERVER, s);
		}
		if (Parent->GetOptions()->Get(OPT_HttpProxyUser, s))
		{
			SetCtrlName(IDC_USER_NAME, s);
		}
		if (Parent->GetOptions()->Get(OPT_HttpProxyPass, s))
		{
			SetCtrlName(IDC_PASSWORD, s);
		}
		if (Parent->GetOptions()->Get(OPT_HttpUrl, s))
		{
			SetCtrlName(IDC_URL, s);
		}
		if (Parent->GetOptions()->Get(OPT_HttpFile, s))
		{
			SetCtrlName(IDC_FILE, s);
			SetResumable(FileExists(s));
		}

		OnNotify(FindControl(IDC_USE_PROXY), 0);
	}
}

HttpDlg::~HttpDlg()
{
	DeleteArray(Url);
	DeleteArray(File);
	DeleteArray(Proxy);
	DeleteArray(UserName);
	DeleteArray(Password);
}

void HttpDlg::OnCreate()
{
	/*
	if (strlen(GetCtrlName(IDC_URL)) > 7)
	{
		PostEvent(WM_CHANGE, (int)FindControl(IDC_BROWSE), 0);
	}
	*/
}

void HttpDlg::SetRollback(bool r)
{
	SetCtrlEnabled(IDC_ROLLBACK, r);

	bool Rolling = GetCtrlValue(IDC_ROLLBACK);
	SetCtrlEnabled(IDC_ROLLBACK_K, r AND Rolling);
	SetCtrlEnabled(IDC_ROLLBACK_STATIC, r AND Rolling);
}

void HttpDlg::SetResumable(bool r)
{
	if (NOT r) SetCtrlValue(IDC_RESUME, 0);
	SetCtrlEnabled(IDC_RESUME_OPT, r);
	SetRollback(GetCtrlValue(IDC_RESUME) == 1);
}

int HttpDlg::OnNotify(GView *Ctrl, int Flags)
{
	if (NOT Ctrl) return 0;

	switch (Ctrl->GetId())
	{
		case IDC_ROLLBACK:
		{
			SetCtrlEnabled(IDC_ROLLBACK_K, Ctrl->Value());
			SetCtrlEnabled(IDC_ROLLBACK_STATIC, Ctrl->Value());
			break;
		}
		case IDC_USE_PROXY:
		{
			int e = GetCtrlValue(IDC_USE_PROXY);
			SetCtrlEnabled(IDC_PROXY_SERVER, e);
			SetCtrlEnabled(IDC_USER_NAME, e);
			SetCtrlEnabled(IDC_PASSWORD, e);
			break;
		}
		case IDC_RESUME:
		{
			SetRollback(GetCtrlValue(IDC_RESUME) == 1);
			break;
		}
		case IDC_BROWSE:
		{
			GFileSelect Select;

			Select.Parent(this);
			Select.Type("All Files", LGI_ALL_FILES);

			char *Url = GetCtrlName(IDC_URL);
			if (Url)
			{
				char *Dir = strrchr(Url, '/');
				if (Dir)
				{
					Select.Name(Dir + 1);
				}
			}

			if (Select.Save())
			{
				SetCtrlName(IDC_FILE, Select.Name());
				SetResumable(FileExists(Select.Name()));
			}
			break;
		}
		case IDOK:
		{
			Url = NewStr(GetCtrlName(IDC_URL));
			File = NewStr(GetCtrlName(IDC_FILE));
			if (GetCtrlValue(IDC_USE_PROXY))
			{
				Proxy = NewStr(GetCtrlName(IDC_PROXY_SERVER));
				UserName = NewStr(GetCtrlName(IDC_USER_NAME));
				Password = NewStr(GetCtrlName(IDC_PASSWORD));
			}
			if (GetCtrlValue(IDC_RESUME) == 1)
			{
				StartFromByte = FileSize(File);
				if (GetCtrlValue(IDC_ROLLBACK))
				{
					int Rollback = GetCtrlValue(IDC_ROLLBACK_K) << 10;
					if (Rollback > 0)
					{
						StartFromByte = max(0, StartFromByte - Rollback);
					}
				}
			}
			NoCache = GetCtrlValue(IDC_NO_CACHE) != 0;
			// fall thru
		}
		case IDCANCEL:
		{
			Parent->GetOptions()->Set(OPT_HttpUrl, GetCtrlName(IDC_URL));
			Parent->GetOptions()->Set(OPT_HttpFile, GetCtrlName(IDC_FILE));

			Parent->GetOptions()->Set(OPT_HttpProxyOn, GetCtrlValue(IDC_USE_PROXY));
			Parent->GetOptions()->Set(OPT_HttpProxy, GetCtrlName(IDC_PROXY_SERVER));
			Parent->GetOptions()->Set(OPT_HttpProxyUser, GetCtrlName(IDC_USER_NAME));
			Parent->GetOptions()->Set(OPT_HttpProxyPass, GetCtrlName(IDC_PASSWORD));

			EndModal(Ctrl->GetId());
			break;
		}
	}

	return 0;
}

//////////////////////////////////////////////////////////////
HttpThread::HttpThread(AppWnd *parent, HttpDlg *dlg)
{
	Parent = parent;
	DeleteOnExit = true;
	Dlg = dlg;

	Run();
}

HttpThread::~HttpThread()
{
	DeleteObj(Dlg);
}

int HttpThread::Main()
{
	if (Parent AND Prog)
	{
		// reset progress
		if (Parent->Progress)
		{
			Parent->Progress->Cancel(false);
		}

		char *HttpStr = "http://";
		if (Dlg AND
			Dlg->Url AND
			Dlg->File)
		{
			ReopenHttp:

			// Parse URL
			char Url[1024];
			int Port = 80;
			char *Path = 0;
			char *FileName = 0;

			strcpy(Url, Dlg->Url);
			char *Server = (strnicmp(Url, HttpStr, strlen(HttpStr)) == 0) ? Url+strlen(HttpStr) : Url;
			if (Server)
			{
				for (char *p = Server+strlen(Server)-1; p>Server; p--)
				{
					if (*p == '/')
					{
						FileName = NewStr(p + 1);
						break;
					}
				}

				char *PathStr = strchr(Server, '/');
				if (PathStr)
				{
					Path = NewStr(PathStr);
					*PathStr = 0;
				}

				char *PortStr = strchr(Server, ':');
				if (PortStr)
				{
					*PortStr++ = 0;
					Port = atoi(PortStr);
				}
			}

			int Format = GET_TYPE_NORMAL;
			DoConnect:

			IHttp Http;
			GSocket *Connection = NEW(HttpLogConnection(Parent->LogPanel, Parent));
			bool OpenOk = false;

			Http.SetNoCache(Dlg->NoCache);
			
			if (Dlg->Proxy)
			{
				char *Col = strchr(Dlg->Proxy, ':');
				if (Col)
				{
					*Col++ = 0;
					Http.SetProxy(Dlg->Proxy, atoi(Col));
					Col[-1] = ':';
				}
				else
				{
					Http.SetProxy(Dlg->Proxy, 80);
				}
			}

			OpenOk = Http.Open(Connection, Server, Port);

			if (OpenOk)
			{
				Http.Meter = Parent->Progress;
				IFtpEntry e;

				e.Name = NewStr(Dlg->File);

				Parent->CmdConnect.Enabled(false);
				Parent->CmdHttp.Enabled(false);
				Parent->CmdAbort.Enabled(true);

				char GetStr[256];
				sprintf(GetStr, "http://%s%s", Server, Path ? Path : (char*)"");
				
				GFile Out;
				if (Out.Open(Dlg->File, O_WRITE))
				{
					if (Dlg->StartFromByte > 0)
					{
						Http.SetResume(Dlg->StartFromByte);

						Out.SetSize(Dlg->StartFromByte);
						Out.Seek(Dlg->StartFromByte, SEEK_SET);
					}
					else
					{
						Out.SetSize(0);
					}

					int ProtocolStatus = 0;
					int ContentLength = 0;
					bool Status = Http.GetFile(0, GetStr, Out, Format, &ProtocolStatus, &ContentLength);

					if (ProtocolStatus == 301 OR
						ProtocolStatus == 302)
					{
						DeleteArray(Dlg->Url);
						Dlg->Url = NewStr(Http.AlternateLocation());
						if (ValidStr(Dlg->Url))
						{
							goto ReopenHttp;
						}
						else
						{
							Status = false;
						}
					}

					if (Status)
					{
						Connection->OnInformation("HTTP download successful.");
					}
					else
					{
						Connection->OnError(-1, "HTTP didn't complete.");

						Out.Close();

						Dlg->StartFromByte = FileSize(Dlg->File);

						if (ProtocolStatus < 0 AND
							Dlg->StartFromByte < ContentLength)
						{
							goto DoConnect;
						}

						if (Format == GET_TYPE_NORMAL)
						{
							Format = GET_TYPE_FILE_HOST;
							goto DoConnect;
						}
					}
				}
				else
				{
					char Str[256];
					sprintf(Str, "Couldn't open '%s' for writing.", Dlg->File);
					Connection->OnInformation(Str);
				}

				Parent->CmdConnect.Enabled(true);
				Parent->CmdHttp.Enabled(true);
				Parent->CmdAbort.Enabled(false);

				if (Parent->Local)
				{
					Parent->Local->ListElements();
				}
			}
			else
			{
				Connection->OnInformation("Couldn't connect to server.");
			}

			DeleteArray(Path);
			DeleteArray(FileName);
		}
		else
		{
			printf("%s:%i - User arguments missing.\n", __FILE__, __LINE__);
		}
	}
	else
	{
		printf("%s:%i - Parameter error.\n", __FILE__, __LINE__);
	}

	if (Parent)
	{
		Parent->Http = 0;
	}
	
	return 0;
}

