#include <time.h>
#include <stdio.h>

#include "Lgi.h"
#include "ISFtp.h"
#include "libssh/libssh.h"
#include "libssh/sftp.h"
#include "GToken.h"
#include "FileTransferProgress.h"
#include "resdefs.h"

char *TerminatePath(char *p)
{
	char *e = 0;
	if (p)
	{
		e = p + strlen(p);
		if (e[-1] != '/')
		{
			*e++ = '/';
			e[1] = 0;
		}
	}
	return e;
}

class libssh : public GLibrary
{
public:
	libssh() : GLibrary("libssh")
	{
	}

	// SSH functions
	GLibFunc1(char *, ssh_get_error, SSH_SESSION *, session);
	GLibFunc3(int, ssh_userauth_password, SSH_SESSION *, session, char *, username, char *, password);
	GLibFunc1(int, ssh_userauth_autopubkey, SSH_SESSION *,session);
	GLibFunc1(char *, ssh_get_issue_banner, SSH_SESSION *, session);
	GLibFunc1(SSH_SESSION *, ssh_connect, SSH_OPTIONS *, options);
	GLibFunc1(int, ssh_is_server_known, SSH_SESSION *, session);
	GLibFunc2(int, options_set_username, SSH_OPTIONS *, opt, char *, username);
	GLibFunc2(int, options_set_port, SSH_OPTIONS *, opt, unsigned int, port);
	GLibFunc2(int, options_set_host, SSH_OPTIONS *, opt, const char *, host);
	GLibFunc0(SSH_OPTIONS *, options_new);
	GLibFunc1(int, ssh_set_verbosity, int, num);
	GLibFunc1(int, ssh_disconnect, SSH_SESSION *, session);
	GLibFunc1(int, ssh_write_knownhost, SSH_SESSION *, session);

	// SFTP functions
	GLibFunc1(int, sftp_init, SFTP_SESSION *,sftp);
	GLibFunc1(SFTP_SESSION *, sftp_new, SSH_SESSION *,session);
	GLibFunc1(int,  sftp_dir_close, SFTP_DIR *,dir);
	GLibFunc2(SFTP_DIR *, sftp_opendir, SFTP_SESSION *,session, char *,path);
	GLibFunc3(int, sftp_mkdir, SFTP_SESSION *,sftp, char *,directory, SFTP_ATTRIBUTES *,attr);
	GLibFunc2(int, sftp_rmdir, SFTP_SESSION *,sftp, char *,directory);
	GLibFunc1(int, sftp_attributes_free, SFTP_ATTRIBUTES *,file);
	GLibFunc2(SFTP_ATTRIBUTES *, sftp_readdir, SFTP_SESSION *,session, SFTP_DIR *,dir);
	GLibFunc2(int, sftp_rm, SFTP_SESSION *,sftp, char *,file);
	GLibFunc1(int, sftp_file_close, SFTP_FILE *,file);
	GLibFunc3(int, sftp_read, SFTP_FILE *,file, void *,dest, int, len);
	GLibFunc3(int, sftp_write, SFTP_FILE *,file, void *,source, int, len);
	GLibFunc2(int, sftp_seek, SFTP_FILE *,file, int, new_offset);
	GLibFunc4(SFTP_FILE *, sftp_open, SFTP_SESSION *,session, char *,file, int, access, SFTP_ATTRIBUTES *,attr);
	GLibFunc3(int, sftp_rename, SFTP_SESSION *,sftp, char *,original, char *,newname);
};

class SFtpPrivate : public libssh
{
public:
	bool ShowHidden;
	Progress *Meter;
	bool Authed;
	SSH_OPTIONS *Opts;
	SSH_SESSION *Sess;
	SFTP_SESSION *Ftp;
	GSocket *Sock;
	char *Path;
	int64 ResumeAt;
	char m[512];
	GView *Parent;
	char *LibName;

	SFtpPrivate()
	{
		ShowHidden = true;
		Authed = false;
		Meter = 0;
		Opts = 0;
		Sess = 0;
		Sock = 0;
		Ftp = 0;
		Parent = 0;
		ResumeAt = 0;
		LibName = NewStr("SFTP");
		Path = NewStr("./");
	}

	~SFtpPrivate()
	{
		DeleteArray(LibName);
	}

	char *MakePath(char *p, char *name)
	{
		char *s;
		while (s = strchr(name, '\\'))
		{
			*s = '/';
		}

		if (strchr(name, '/'))
		{
			strcpy(p, name);
		}
		else if (Path)
		{
			strcpy(p, Path);
			char *e = TerminatePath(p);
			strcpy(e, name);
		}
		return p;
	}
};

SFtp::SFtp(GView *Parent)
{
	d = NEW(SFtpPrivate);
	d->Parent = Parent;
}

SFtp::~SFtp()
{
	DeleteObj(d);
}

bool SFtp::IsForceActive()
{
	return false;
}

void SFtp::IsForceActive(bool i)
{
}

bool SFtp::IsLongList()
{
	return false;
}

void SFtp::IsLongList(bool i)
{
}

bool SFtp::IsShowHidden()
{
	return d->ShowHidden;
}

void SFtp::IsShowHidden(bool i)
{
	d->ShowHidden = i;
}

Progress *SFtp::GetMeter()
{
	return d->Meter;
}

void SFtp::SetMeter(Progress *m)
{
	d->Meter = m;
}

bool SFtp::GetAuthed()
{
	return d->Authed;
}

GSocket *SFtp::Handle()
{
	return d->Sock;
}

FtpOpenStatus SFtp::Open(GSocket *S, char *RemoteHost, int Port, char *User, char *Password)
{
	FtpOpenStatus Status = FO_ConnectFailed;
	
	Close();

	if (NOT d->IsLoaded())
	{
		if (S)
		{
			S->OnError(0, "libssh[.dll] (or OpenSSL libraries) missing.");
		}
		Status = FO_Error;
	}
	else if (RemoteHost AND User AND Password)
	{
		// d->ssh_set_verbosity(3);

		d->Sock = S;
		if (d->Opts = d->options_new())
		{
			d->options_set_host(d->Opts, RemoteHost);
			d->options_set_port(d->Opts, Port);
			d->options_set_username(d->Opts, User);

			if (d->Sess = d->ssh_connect(d->Opts))
			{
				char AuthFile[300];
				if (LgiGetSystemPath(LSP_APP_DATA, AuthFile, sizeof(AuthFile)))
				{
					LgiMakePath(AuthFile, sizeof(AuthFile), AuthFile, ".ssh");
					if (NOT DirExists(AuthFile))
					{
						FileDev->CreateDirectory(AuthFile);
					}

					LgiMakePath(AuthFile, sizeof(AuthFile), AuthFile, "known_hosts");
					if (NOT FileExists(AuthFile))
					{
						GFile f;
						f.Open(AuthFile, O_WRITE);
					}
				}

				bool Authed = false;
			    int State = d->ssh_is_server_known(d->Sess);

				switch (State)
				{
					case SSH_SERVER_KNOWN_OK:
					{
						Authed = true;
						break;
					}
					case SSH_SERVER_KNOWN_CHANGED:
					{
						GAlert a(	d->Parent,
									d->LibName,
									LgiLoadString(IDS_SFTP_HOST_CHANGED),
									LgiLoadString(IDS_SFTP_ACCEPT_CHANGE),
									LgiLoadString(IDS_CANCEL));
						switch (a.DoModal())
						{
							case 1:
							{
								Authed = true;
								d->ssh_write_knownhost(d->Sess);
								break;
							}
							default:
							{
								Status = FO_Error;
								break;
							}
						}
						break;
					}
					case SSH_SERVER_FOUND_OTHER:
					{
						GAlert a(	d->Parent,
									d->LibName,
									LgiLoadString(IDS_SFTP_FOUND_OTHER),
									LgiLoadString(IDS_SFTP_ACCEPT_CHANGE),
									LgiLoadString(IDS_CANCEL));
						switch (a.DoModal())
						{
							case 1:
							{
								Authed = true;
								d->ssh_write_knownhost(d->Sess);
								break;
							}
							default:
							{
								Status = FO_Error;
								break;
							}
						}
						break;
					}
					case SSH_SERVER_NOT_KNOWN:
					{
						int Result = LgiMsg(d->Parent, LgiLoadString(IDS_SFTP_UNKNOWN_SERVER), d->LibName, MB_YESNO);
						if (Result == IDYES)
						{
							Authed = true;
							d->ssh_write_knownhost(d->Sess);
						}
						else
						{
							Status = FO_Error;
						}
						break;
					}
					case SSH_SERVER_ERROR:
					{
						d->Sock->OnError(0, d->ssh_get_error(d->Sess));
						Status = FO_Error;
						break;
					}
				}

				if (Authed)
				{
					d->ssh_userauth_autopubkey(d->Sess);

					GToken Banner(d->ssh_get_issue_banner(d->Sess), "\r\n");
					for (int i=0; i<Banner.Length(); i++)
					{
						d->Sock->OnRead(Banner[i], strlen(Banner[i]));
					}

					if (d->ssh_userauth_password(d->Sess, 0, Password) == SSH_AUTH_SUCCESS)
					{
						if ((d->Ftp = d->sftp_new(d->Sess)) AND
							d->sftp_init(d->Ftp) == 0)
						{
							d->Sock->OnInformation("Connected...");
							Status = FO_Connected;
						}
						else
						{
							d->Sock->OnError(0, d->ssh_get_error(d->Sess));
							Status = FO_Error;
						}
					}
					else
					{
						d->Sock->OnError(0, d->ssh_get_error(d->Sess));
						Status = FO_LoginFailed;
						Close();
					}
				}
				else
				{
					Close();
				}
			}
			else
			{
				d->Sock->OnError(0, d->ssh_get_error(0));
			}
		}
	}

	return Status;
}

bool SFtp::Close()
{
	if (d->Sess)
	{
		d->ssh_disconnect(d->Sess);
		d->Sess = 0;
	}

	d->Opts = 0;

	return true;
}

bool SFtp::IsOpen()
{
	return d->Sess != 0;
}

void SFtp::Noop()
{
}

bool SFtp::GetDir(char *Dir)
{
	if (Dir AND d->Path)
	{
		strcpy(Dir, d->Path);
		return true;
	}
	return false;
}

bool SFtp::SetDir(char *Dir)
{
	bool Status = false;

	if (Dir AND d->Path)
	{
		char p[512];
		d->MakePath(p, Dir);

		sprintf(d->m, "CD %s\n", p);
		d->Sock->OnWrite(d->m, strlen(d->m));

		SFTP_DIR *f = d->sftp_opendir(d->Ftp, p);
		if (f)
		{
			DeleteArray(d->Path);
			d->Path = NewStr(p);
			Status = true;
			d->Sock->OnRead("Ok\n", 3);

			d->sftp_dir_close(f);
		}
		else
		{
			d->Sock->OnError(0, d->ssh_get_error(d->Sess));
		}
	}
	else
	{
		d->Sock->OnError(0, "Invalid parameters.");
	}

	return Status;
}

bool SFtp::CreateDir(char *Dir)
{
	bool Status = false;

	if (Dir AND d->Ftp)
	{
		char p[512] = "";
		d->MakePath(p, Dir);

		sprintf(d->m, "MKDIR %s\n", p);
		d->Sock->OnWrite(d->m, strlen(d->m));

		SFTP_ATTRIBUTES a;
		ZeroObj(a);
		if (d->sftp_mkdir(d->Ftp, p, &a) == 0)
		{
			d->Sock->OnRead("Ok\n", 3);
			Status = true;
		}
		else
		{
			d->Sock->OnError(0, d->ssh_get_error(d->Sess));
		}
	}

	return Status;
}

bool SFtp::DeleteDir(char *Dir)
{
	bool Status = false;

	if (Dir AND d->Ftp)
	{
		char p[512] = "";
		d->MakePath(p, Dir);

		sprintf(d->m, "RMDIR %s\n", p);
		d->Sock->OnWrite(d->m, strlen(d->m));

		if (d->sftp_rmdir(d->Ftp, p) == 0)
		{
			d->Sock->OnRead("Ok\n", 3);
			Status = true;
		}
		else
		{
			d->Sock->OnError(0, d->ssh_get_error(d->Sess));
		}
	}

	return Status;
}

void ConvertDate(GDateTime &Out, uint32 In)
{
	time_t in = In;
	struct tm *m = localtime(&in);
	if (m)
	{
		Out.Year(m->tm_year + 1900);
		Out.Month(m->tm_mon + 1);
		Out.Day(m->tm_mday);
		Out.Hours(m->tm_hour);
		Out.Minutes(m->tm_min);
		Out.Seconds(m->tm_sec);
	}
}

bool SFtp::ListDir(List<IFtpEntry> *Dir)
{
	bool Status = false;

	if (Dir AND d->Sess)
	{
		SFTP_DIR *f = d->sftp_opendir(d->Ftp, d->Path);
		if (f)
		{
			SFTP_ATTRIBUTES *a;
			while (a = d->sftp_readdir(d->Ftp, f))
			{
				if (strcmp(a->name, ".") AND
					strcmp(a->name, ".."))
				{
					IFtpEntry *e = NEW(IFtpEntry);
					if (e)
					{
						e->Size = a->size;
						e->Name = NewStr(a->name);
						e->Path = NewStr(d->Path);
						e->Attributes = (a->longname AND *a->longname == 'd') ? IFTP_DIR : 0;
						ConvertDate(e->Date, a->mtime);

						Dir->Insert(e);
					}
				}

				d->sftp_attributes_free(a);
			}

			Status = true;
			d->sftp_dir_close(f);
		}
		else
		{
			d->Sock->OnError(0, d->ssh_get_error(d->Sess));
		}
	}
	
	return Status;
}

bool SFtp::UpDir()
{
	bool Status = false;

	if (d->Path)
	{
		char *p = strrchr(d->Path, '/');
		if (p)
		{
			if (p - d->Path <= 3)
			{
				p[1] = 0;
			}
			else
			{
				*p = 0;
			}
			Status = true;
		}
		else
		{
			LgiAssert(0);
		}
	}

	return Status;
}

bool SFtp::DeleteFile(char *Remote)
{
	bool Status = false;

	if (Remote AND d->Ftp)
	{
		char p[512] = "";
		d->MakePath(p, Remote);

		sprintf(d->m, "DEL %s\n", p);
		d->Sock->OnWrite(d->m, strlen(d->m));

		if (d->sftp_rm(d->Ftp, p) == 0)
		{
			d->Sock->OnRead("Ok\n", 3);
			Status = true;
		}
		else
		{
			d->Sock->OnError(0, d->ssh_get_error(d->Sess));
		}
	}

	return Status;
}

bool SFtp::DownloadFile(char *Local, IFtpEntry *Remote, bool Binary)
{
	bool Status = false;

	if (Local AND Remote AND d->Ftp)
	{
		char p[512];
		d->MakePath(p, Remote->Name);

		sprintf(d->m, "GET %s\n", p);
		d->Sock->OnWrite(d->m, strlen(d->m));

		SFTP_FILE *In = d->sftp_open(d->Ftp, p, O_RDONLY, 0);
		if (In)
		{
			Status = true;

			GFile Out;
			if (Out.Open(Local, O_WRITE))
			{
				d->Sock->OnRead("Ok", 3);
				Out.SetSize(d->ResumeAt);
				Out.SetPos(d->ResumeAt);
				
				int BufLen = 32 << 10;
				char *Buf = NEW(char[BufLen]);
				if (Buf)
				{
					if (d->Meter)
					{
						d->Meter->SetLimits(0, Remote->Size-1);
					}

					if (d->ResumeAt)
					{
						d->sftp_seek(In, d->ResumeAt);
						if (d->Meter)
						{
							d->Meter->Value(d->ResumeAt);
							d->Meter->SetParameter(PARM_START_VALUE, d->ResumeAt);
						}
					}

					int r;
					while ((r = d->sftp_read(In, Buf, BufLen)) > 0)
					{
						Out.Write(Buf, r);
						if (d->Meter)
						{
							d->Meter->Value(d->Meter->Value() + r);
						}
					}

					if (d->Meter)
					{
						d->Meter->Value(0);
						d->Meter->SetLimits(0, 0);
					}

					DeleteArray(Buf);
				}

				d->ResumeAt = 0;
			}
			else
			{
				sprintf(d->m, "Can't open '%s' for writing.\n", Local);
				d->Sock->OnError(0, d->m);
			}

			d->sftp_file_close(In);
		}
		else
		{
			d->Sock->OnError(0, d->ssh_get_error(d->Sess));
		}
	}

	return Status;
}

bool SFtp::UploadFile(char *Local, char *Remote, bool Binary)
{
	bool Status = false;

	if (Local AND Remote AND d->Ftp)
	{
		GFile In;
		if (In.Open(Local, O_READ))
		{
			char p[512];
			d->MakePath(p, Remote);

			sprintf(d->m, "PUT %s\n", p);
			d->Sock->OnWrite(d->m, strlen(d->m));

			SFTP_FILE *Out = d->sftp_open(d->Ftp, p, O_WRONLY | O_CREAT, 0);
			if (Out)
			{
				d->Sock->OnRead("Ok", 3);
				Status = true;

				int BufLen = 32 << 10;
				char *Buf = NEW(char[BufLen]);
				if (Buf)
				{
					if (d->Meter)
					{
						d->Meter->SetLimits(0, FileSize(Local)-1);
					}

					int r;
					while ((r = In.Read(Buf, BufLen)) > 0)
					{
						int w = d->sftp_write(Out, Buf, r);
						if (w == r)
						{
							if (d->Meter)
							{
								d->Meter->Value(d->Meter->Value() + r);
							}
						}
						else
						{
							d->Sock->OnError(0, d->ssh_get_error(d->Sess));
							break;
						}
					}

					if (d->Meter)
					{
						d->Meter->Value(0);
						d->Meter->SetLimits(0, 0);
					}

					DeleteArray(Buf);
				}

				d->sftp_file_close(Out);
			}
			else
			{
				d->Sock->OnError(0, d->ssh_get_error(d->Sess));
			}
		}
		else
		{
			sprintf(d->m, "Can't open '%s' for reading\n", Local);
			d->Sock->OnError(0, d->m);
		}
	}

	return Status;
}

bool SFtp::RenameFile(char *From, char *To)
{
	bool Status = false;

	if (From AND To AND d->Ftp)
	{
		char p[512] = "";
		d->MakePath(p, From);
		char to[512] = "";
		d->MakePath(to, To);

		sprintf(d->m, "RENAME '%s' -> '%s'\n", p, to);
		d->Sock->OnWrite(d->m, strlen(d->m));

		if (d->sftp_rename(d->Ftp, p, to) == 0)
		{
			d->Sock->OnRead("Ok\n", 3);
			Status = true;
		}
		else
		{
			d->Sock->OnError(0, d->ssh_get_error(d->Sess));
		}
	}

	return Status;
}

bool SFtp::SetPerms(char *File, int Perms)
{
	LgiAssert(0);
	return false;
}

bool SFtp::ResumeAt(int Pos)
{
	d->ResumeAt = Pos;
	return true;
}

void SFtp::Abort()
{
	LgiAssert(0);
}

